import tkinter as tk
from tkinter import ttk

from thonny.config_ui import ConfigurationPage
from thonny.globals import get_workbench


class GeneralConfigurationPage(ConfigurationPage):
    
    def __init__(self, master):
        ConfigurationPage.__init__(self, master)
        
        self._single_instance_var = get_workbench().get_variable("general.single_instance")
        self._single_instance_checkbox = ttk.Checkbutton(self,
                                                         text="Allow only single Thonny instance", 
                                                      variable=self._single_instance_var)
        self._single_instance_checkbox.grid(row=1, column=0, sticky=tk.W)
        
        self._expert_var = get_workbench().get_variable("general.expert_mode")
        self._expert_checkbox = ttk.Checkbutton(self, text="Expert mode", variable=self._expert_var)
        self._expert_checkbox.grid(row=2, column=0, sticky=tk.W)
        
        self._debug_var = get_workbench().get_variable("general.debug_mode")
        self._debug_checkbox = ttk.Checkbutton(self, text="Debug mode", variable=self._debug_var)
        self._debug_checkbox.grid(row=3, column=0, sticky=tk.W)
        
        reopen_label = ttk.Label(self, text="NB! Restart Thonny after changing these options"
                                 + "\nin order to see the effect")
        reopen_label.grid(row=4, column=0, sticky=tk.W, pady=20)
        
        
        self.columnconfigure(0, weight=1)

            
    

def load_plugin():
    get_workbench().add_configuration_page("General", GeneralConfigurationPage)



# -*- coding: utf-8 -*-
from tkinter import ttk
from thonny.misc_utils import running_on_linux
from thonny.globals import get_workbench

_images = set() # for keeping references to tkinter images to avoid garbace colleting them



def tweak_notebooks():
    style = ttk.Style()
    theme = style.theme_use()
    
    if theme in ["xpnative", "vista"]:
        get_workbench().get_image('gray_line.gif', "gray_line")
        
        style.element_create("gray_line", "image", "gray_line",
                                   ("!selected", "gray_line"), 
                                   height=1, width=10, border=1)
        
        style.layout('Tab', [
            ('Notebook.tab', {'sticky': 'nswe', 'children': [
                ('Notebook.padding', {'sticky': 'nswe', 'side': 'top', 'children': [
                    ('Notebook.focus', {'sticky': 'nswe', 'side': 'top', 'children': [
                        ('Notebook.label', {'sticky': '', 'side': 'left'}),
                    ]})
                ]}),
                ('gray_line', {'sticky': 'we', 'side': 'bottom'}),
            ]}),
        ])
    
    style.configure("Tab", padding=(4,1,0,0))
    if theme == "clam":
        style.configure("ButtonNotebook.Tab", padding=(6,4,2,3))
    else:
        style.configure("ButtonNotebook.Tab", padding=(4,1,1,3))
            
    if theme == "aqua":
        style.map("TNotebook.Tab", foreground=[('selected', 'white'), ('!selected', 'black')])


def tweak_treeviews():
    style = ttk.Style()
    # get rid of Treeview borders
    style.layout("Treeview", [
        ('Treeview.treearea', {'sticky': 'nswe'})
    ])
    
    # necessary for Python 2.7 TODO: doesn't help for aqua
    style.configure("Treeview", background="white")
    
    #style.configure("Treeview", font='helvetica 14 bold')
    style.configure("Treeview", font=get_workbench().get_font("TreeviewFont"))

    #print(style.map("Treeview"))
    #print(style.layout("Treeview"))
    #style.configure("Treeview.treearea", font=TREE_FONT)
    # NB! Some Python or Tk versions (Eg. Py 3.2.3 + Tk 8.5.11 on Raspbian)
    # can't handle multi word color names in style.map  
    light_blue = "#ADD8E6" 
    light_grey = "#D3D3D3"
    if running_on_linux():
        style.map("Treeview",
              background=[('selected', 'focus', light_blue),
                          ('selected', '!focus', light_grey),
                          ],
              foreground=[('selected', 'black'),
                          ],
              )
    else:
        style.map("Treeview",
              background=[('selected', 'focus', 'SystemHighlight'),
                          ('selected', '!focus', light_grey),
                          ],
              foreground=[('selected', 'SystemHighlightText')],
              )

def tweak_menubuttons():
    style = ttk.Style()
    #print(style.layout("TMenubutton"))
    style.layout("TMenubutton", [
        ('Menubutton.dropdown', {'side': 'right', 'sticky': 'ns'}),
        ('Menubutton.button', {'children': [
            #('Menubutton.padding', {'children': [
                ('Menubutton.label', {'sticky': ''})
            #], 'expand': '1', 'sticky': 'we'})
        ], 'expand': '1', 'sticky': 'nswe'})
    ])
    
    style.configure("TMenubutton", padding=14)

def tweak_paned_windows():
    style = ttk.Style()
    style.configure("Sash", sashthickness=10)


def load_plugin():
    tweak_notebooks()
    tweak_treeviews()
    tweak_paned_windows()
    
    



# -*- coding: utf-8 -*-
"""
This file is run by VMProxy

(Why separate file for launching? I want to have clean global scope 
in toplevel __main__ module (because that's where user scripts run), but backend's global scope 
is far from clean. 
I could also do python -c "from backend import VM: VM().mainloop()", but looks like this 
gives relative __file__-s on imported modules.) 
"""

if __name__ == "__main__":
    # imports required by backend itself
    import sys
    import logging
    import os.path
    
    # set up logging
    logger = logging.getLogger()
    logFormatter = logging.Formatter('%(levelname)s: %(message)s')
    file_handler = logging.FileHandler(os.path.expanduser(os.path.join(os.environ["THONNY_USER_DIR"],"backend.log")), 
                                       encoding="UTF-8",
                                       mode="w");
    file_handler.setFormatter(logFormatter)
    file_handler.setLevel(logging.INFO);
    logger.addHandler(file_handler)
    
    # TODO: sending log records to original stdout could be better (reading from stderr may introduce sync problems)
    stream_handler = logging.StreamHandler(stream=sys.stderr)
    stream_handler.setLevel(logging.INFO);
    stream_handler.setFormatter(logFormatter)
    logger.addHandler(stream_handler)
    
    logger.setLevel(logging.INFO)
    
    from thonny.backend import VM  # @UnresolvedImport
    VM().mainloop()
    



# -*- coding: utf-8 -*-

import tkinter as tk

from thonny.memory import MemoryFrame, format_object_id, MAX_REPR_LENGTH_IN_GRID,\
    parse_object_id
from thonny.misc_utils import shorten_repr
from thonny.globals import get_workbench, get_runner
from thonny.common import InlineCommand

class HeapView(MemoryFrame):
    def __init__(self, master):
        MemoryFrame.__init__(self, master, ("id", "value"))
        
        self.tree.column('id', width=100, anchor=tk.W, stretch=False)
        self.tree.column('value', width=150, anchor=tk.W, stretch=True)
        
        self.tree.heading('id', text='ID', anchor=tk.W)
        self.tree.heading('value', text='Value', anchor=tk.W) 
        
        get_workbench().bind("Heap", self._handle_heap_event, True)
        
        get_workbench().bind("DebuggerProgress", self._request_heap_data, True)
        get_workbench().bind("ToplevelResult", self._request_heap_data, True)
        # Showing new globals may introduce new interesting objects
        get_workbench().bind("Globals", self._request_heap_data, True) 

    def _update_data(self, data):
        self._clear_tree()
        for value_id in sorted(data.keys()):
            node_id = self.tree.insert("", "end")
            self.tree.set(node_id, "id", format_object_id(value_id))
            self.tree.set(node_id, "value", shorten_repr(data[value_id]["repr"], MAX_REPR_LENGTH_IN_GRID))
    
    def before_show(self):
        self._request_heap_data(even_when_hidden=True)

    def on_select(self, event):
        iid = self.tree.focus()
        if iid != '':
            object_id = parse_object_id(self.tree.item(iid)['values'][0])
            get_workbench().event_generate("ObjectSelect", object_id=object_id)
            
    def _request_heap_data(self, msg=None, even_when_hidden=False):
        if self.winfo_ismapped() or even_when_hidden:
            # TODO: update itself also when it becomes visible
            get_runner().send_command(InlineCommand("get_heap"))
            
    def _handle_heap_event(self, msg):
        if self.winfo_ismapped():
            if hasattr(msg, "heap"):
                self._update_data(msg.heap)
                
def load_plugin():
    get_workbench().add_view(HeapView, "Heap", "e")


# -*- coding: utf-8 -*-
"""Adds commands for opening certain Thonny folders"""

import os.path
from thonny.misc_utils import running_on_mac_os, running_on_linux,\
    running_on_windows
import subprocess
from thonny.globals import get_workbench
from thonny import THONNY_USER_DIR

def open_path_in_system_file_manager(path):
    if running_on_mac_os():
        # http://stackoverflow.com/a/3520693/261181
        # -R doesn't allow showing hidden folders
        subprocess.Popen(["open", path])
    elif running_on_linux():
        subprocess.Popen(["xdg-open", path])
    else:
        assert running_on_windows()
        subprocess.Popen(["explorer", path])




def load_plugin():
    def cmd_open_data_dir():
        open_path_in_system_file_manager(THONNY_USER_DIR)
        
    def cmd_open_program_dir():
        open_path_in_system_file_manager(get_workbench().get_package_dir())
        
    get_workbench().add_command("open_program_dir", "tools", "Open Thonny program folder...",
                                cmd_open_program_dir, group=110)
    get_workbench().add_command("open_data_dir", "tools", "Open Thonny data folder...",
                                cmd_open_data_dir, group=110)
    


# -*- coding: utf-8 -*-

import tkinter as tk
from thonny.common import TextRange
from thonny.globals import get_workbench
from thonny import tktextext, roughparse
from thonny.ui_utils import EnhancedTextWithLogging
from thonny.tktextext import EnhancedText

EDIT_BACKGROUND="white"
READ_ONLY_BACKGROUND="LightYellow"

class PythonText(EnhancedText):
    
    def perform_return(self, event):
        # copied from idlelib.EditorWindow (Python 3.4.2)
        # slightly modified
        
        text = event.widget
        assert text is self
        
        try:
            # delete selection
            first, last = text.get_selection_indices()
            if first and last:
                text.delete(first, last)
                text.mark_set("insert", first)
            
            # Strip whitespace after insert point
            # (ie. don't carry whitespace from the right of the cursor over to the new line)
            while text.get("insert") in [" ", "\t"]:
                text.delete("insert")
            
            left_part = text.get("insert linestart", "insert")
            # locate first non-white character
            i = 0
            n = len(left_part)
            while i < n and left_part[i] in " \t":
                i = i+1
            
            # is it only whitespace?
            if i == n:
                # start the new line with the same whitespace
                text.insert("insert", '\n' + left_part)
                return "break"
            
            # Turned out the left part contains visible chars
            # Remember the indent
            indent = left_part[:i]
            
            # Strip whitespace before insert point
            # (ie. after inserting the linebreak this line doesn't have trailing whitespace)
            while text.get("insert-1c", "insert") in [" ", "\t"]:
                text.delete("insert-1c", "insert")
                
            # start new line
            text.insert("insert", '\n')
    
            # adjust indentation for continuations and block
            # open/close first need to find the last stmt
            lno = tktextext.index2line(text.index('insert'))
            y = roughparse.RoughParser(text.indentwidth, text.tabwidth)
            
            for context in roughparse.NUM_CONTEXT_LINES:
                startat = max(lno - context, 1)
                startatindex = repr(startat) + ".0"
                rawtext = text.get(startatindex, "insert")
                y.set_str(rawtext)
                bod = y.find_good_parse_start(
                          False,
                          roughparse._build_char_in_string_func(startatindex))
                if bod is not None or startat == 1:
                    break
            y.set_lo(bod or 0)
    
            c = y.get_continuation_type()
            if c != roughparse.C_NONE:
                # The current stmt hasn't ended yet.
                if c == roughparse.C_STRING_FIRST_LINE:
                    # after the first line of a string; do not indent at all
                    pass
                elif c == roughparse.C_STRING_NEXT_LINES:
                    # inside a string which started before this line;
                    # just mimic the current indent
                    text.insert("insert", indent)
                elif c == roughparse.C_BRACKET:
                    # line up with the first (if any) element of the
                    # last open bracket structure; else indent one
                    # level beyond the indent of the line with the
                    # last open bracket
                    text._reindent_to(y.compute_bracket_indent())
                elif c == roughparse.C_BACKSLASH:
                    # if more than one line in this stmt already, just
                    # mimic the current indent; else if initial line
                    # has a start on an assignment stmt, indent to
                    # beyond leftmost =; else to beyond first chunk of
                    # non-whitespace on initial line
                    if y.get_num_lines_in_stmt() > 1:
                        text.insert("insert", indent)
                    else:
                        text._reindent_to(y.compute_backslash_indent())
                else:
                    assert 0, "bogus continuation type %r" % (c,)
                return "break"
    
            # This line starts a brand new stmt; indent relative to
            # indentation of initial line of closest preceding
            # interesting stmt.
            indent = y.get_base_indent_string()
            text.insert("insert", indent)
            if y.is_block_opener():
                text.perform_smart_tab(event)
            elif indent and y.is_block_closer():
                text.perform_smart_backspace(event)
            return "break"
        finally:
            text.see("insert")
            text.event_generate("<<NewLine>>")
            return "break" 


class CodeViewText(EnhancedTextWithLogging, PythonText):
    """Provides opportunities for monkey-patching by plugins"""
    def __init__(self, master=None, cnf={}, **kw):
        if not "background" in kw:
            kw["background"] = EDIT_BACKGROUND
        
        EnhancedTextWithLogging.__init__(self, master=master, cnf=cnf, **kw)
        self._original_background = kw["background"]
        # Allow binding to events of all CodeView texts
        self.bindtags(self.bindtags() + ('CodeViewText',))
        tktextext.fixwordbreaks(tk._default_root)
    
    def set_read_only(self, value):
        EnhancedTextWithLogging.set_read_only(self, value)
        if value:
            self.configure(background=READ_ONLY_BACKGROUND)
        else:
            self.configure(background=self._original_background)

    
    def on_secondary_click(self, event):
        super().on_secondary_click(event)
        get_workbench().get_menu("edit").tk_popup(event.x_root, event.y_root)

class CodeView(tktextext.TextFrame):
    def __init__(self, master, propose_remove_line_numbers=False, **text_frame_args):
        tktextext.TextFrame.__init__(self, master, text_class=CodeViewText,
                                     undo=True, wrap=tk.NONE, background=EDIT_BACKGROUND,
                                     **text_frame_args)
        
        # TODO: propose_remove_line_numbers on paste??
        
        self.text.bind("<<TextChange>>", self._on_text_changed, True)
        
    def get_content(self):
        return self.text.get("1.0", "end-1c") # -1c because Text always adds a newline itself
    
    def set_content(self, content):
        self.text.direct_delete("1.0", tk.END)
        self.text.direct_insert("1.0", content)
        self.update_line_numbers()
        self.text.edit_reset();

        self.text.event_generate("<<TextChange>>")
    
    def _on_text_changed(self, event):
        self.update_line_numbers()
        self.update_margin_line()
    
    def select_lines(self, first_line, last_line):
        self.text.tag_remove("sel", "1.0", tk.END)
        self.text.tag_add("sel", "%s.0" % first_line, "%s.end" % last_line)
    
    def select_range(self, text_range):
        self.text.tag_remove("sel", "1.0", tk.END)
        
        if text_range:
            if isinstance(text_range, int):
                # it's line number
                start = str(text_range - self._first_line_number + 1) + ".0"
                end = str(text_range - self._first_line_number + 1) + ".end"
            elif isinstance(text_range, TextRange):
                start = "%s.%s" % (text_range.lineno - self._first_line_number + 1, text_range.col_offset)
                end = "%s.%s" % (text_range.end_lineno - self._first_line_number + 1, text_range.end_col_offset)
            else:
                assert isinstance(text_range, tuple)
                start, end  = text_range
                
            self.text.tag_add("sel", start, end)
            if isinstance(text_range, int):
                self.text.mark_set("insert", end) 
            self.text.see("%s -1 lines" % start)
            
    
    def get_selected_range(self):
        if self.text.has_selection():
            lineno, col_offset = map(int, self.text.index(tk.SEL_FIRST).split("."))
            end_lineno, end_col_offset = map(int, self.text.index(tk.SEL_LAST).split("."))
        else:
            lineno, col_offset = map(int, self.text.index(tk.INSERT).split("."))
            end_lineno, end_col_offset = lineno, col_offset
            
        return TextRange(lineno, col_offset, end_lineno, end_col_offset)



# -*- coding: utf-8 -*-

import tkinter as tk

from thonny.memory import format_object_id, VariablesFrame, MemoryFrame,\
    MAX_REPR_LENGTH_IN_GRID
from thonny.misc_utils import shorten_repr
from thonny.ui_utils import ScrollableFrame, CALM_WHITE, update_entry_text
from thonny.tktextext import TextFrame
from thonny.common import InlineCommand
import ast
from thonny.globals import get_workbench, get_runner
from logging import exception



class AttributesFrame(VariablesFrame):
    def __init__(self, master):
        VariablesFrame.__init__(self, master)
        self.configure(border=1)
        self.vert_scrollbar.grid_remove()
       
    def on_select(self, event):
        pass
    
    def on_double_click(self, event):
        self.show_selected_object_info()
    


        
    

class ObjectInspector(ScrollableFrame):
    def __init__(self, master):
        
        ScrollableFrame.__init__(self, master)
        
        self.object_id = None
        self.object_info = None
        get_workbench().bind("ObjectSelect", self.show_object, True)
        
        self.grid_frame = tk.Frame(self.interior, bg=CALM_WHITE) 
        self.grid_frame.grid(row=0, column=0, sticky=tk.NSEW, padx=(10,0), pady=15)
        self.grid_frame.columnconfigure(1, weight=1)
        
        def _add_main_attribute(row, caption):
            label = tk.Label(self.grid_frame, text=caption + ":  ",
                             background=CALM_WHITE,
                             justify=tk.LEFT)
            label.grid(row=row, column=0, sticky=tk.NW)
            
            value = tk.Entry(self.grid_frame,
                             background=CALM_WHITE,
                             bd=0,
                             readonlybackground=CALM_WHITE,
                             highlightthickness = 0,
                             state="readonly"
                             )
            if row > 0:
                value.grid(row=row, column=1, columnspan=3, 
                       sticky=tk.NSEW, pady=2)
            else:
                value.grid(row=row, column=1, sticky=tk.NSEW, pady=2)
            return value
        
        self.id_entry   = _add_main_attribute(0, "id")
        self.repr_entry = _add_main_attribute(1, "repr")
        self.type_entry = _add_main_attribute(2, "type")
        self.type_entry.config(cursor="hand2", fg="dark blue")
        self.type_entry.bind("<Button-1>", self.goto_type)
        
        self._add_block_label(5, "Attributes")
        self.attributes_frame = AttributesFrame(self.grid_frame)
        self.attributes_frame.grid(row=6, column=0, columnspan=4, sticky=tk.NSEW, padx=(0,10))
        
        self.grid_frame.grid_remove()
        
        # navigation 
        self.back_label = self.create_navigation_link(2, " << ", self.navigate_back)
        self.forward_label = self.create_navigation_link(3, " >> ", self.navigate_forward, (0,10))
        self.back_links = []
        self.forward_links = []
        
        # type-specific inspectors
        self.current_type_specific_inspector = None
        self.current_type_specific_label = None
        self.type_specific_inspectors = [ 
            FileHandleInspector(self.grid_frame),
            FunctionInspector(self.grid_frame),
            StringInspector(self.grid_frame),
            ElementsInspector(self.grid_frame),
            DictInspector(self.grid_frame),
        ]

        get_workbench().bind("ObjectInfo", self._handle_object_info_event, True)
        get_workbench().bind("DebuggerProgress", self._handle_progress_event, True)
        get_workbench().bind("ToplevelResult", self._handle_progress_event, True)


    
    def create_navigation_link(self, col, text, action, padx=0):
        link = tk.Label(self.grid_frame,
                        text=text,
                        background=CALM_WHITE,
                        foreground="blue",
                        cursor="hand2")
        link.grid(row=0, column=col, sticky=tk.NE, padx=padx)
        link.bind("<Button-1>", action)
        return link
    
    def navigate_back(self, event):
        if len(self.back_links) == 0:
            return
        
        self.forward_links.append(self.object_id)
        self._show_object_by_id(self.back_links.pop(), True)
    
    def navigate_forward(self, event):
        if len(self.forward_links) == 0:
            return
    
        self.back_links.append(self.object_id)
        self._show_object_by_id(self.forward_links.pop(), True)

    def show_object(self, event):
        self._show_object_by_id(event.object_id)
        
    def _show_object_by_id(self, object_id, via_navigation=False):
        
        if self.winfo_ismapped() and self.object_id != object_id:
            if not via_navigation and self.object_id is not None:
                if self.object_id in self.back_links:
                    self.back_links.remove(self.object_id)
                self.back_links.append(self.object_id)
                del self.forward_links[:]
                
            self.object_id = object_id
            update_entry_text(self.id_entry, format_object_id(object_id))
            self.set_object_info(None)
            self.request_object_info()
    
    def _handle_object_info_event(self, msg):
        if self.winfo_ismapped():
            if msg.info["id"] == self.object_id:
                if hasattr(msg, "not_found") and msg.not_found:
                    self.object_id = None
                    self.set_object_info(None)
                else:
                    self.set_object_info(msg.info)
    
    def _handle_progress_event(self, event):
        if self.object_id is not None:
            self.request_object_info()
                
                
    def request_object_info(self): 
        get_runner().send_command(InlineCommand("get_object_info",
                                            object_id=self.object_id,
                                            all_attributes=False)) 
                    
    def set_object_info(self, object_info):
        self.object_info = object_info
        if object_info is None:
            update_entry_text(self.repr_entry, "")
            update_entry_text(self.type_entry, "")
            self.grid_frame.grid_remove()
        else:
            update_entry_text(self.repr_entry, object_info["repr"])
            update_entry_text(self.type_entry, object_info["type"])
            self.attributes_frame.tree.configure(height=len(object_info["attributes"]))
            self.attributes_frame.update_variables(object_info["attributes"])
            self.update_type_specific_info(object_info)
                
            
            # update layout
            self._expose(None)
            if not self.grid_frame.winfo_ismapped():
                self.grid_frame.grid()
    
        if self.back_links == []:
            self.back_label.config(foreground="lightgray", cursor="arrow")
        else:
            self.back_label.config(foreground="blue", cursor="hand2")
    
        if self.forward_links == []:
            self.forward_label.config(foreground="lightgray", cursor="arrow")
        else:
            self.forward_label.config(foreground="blue", cursor="hand2")
    
    def update_type_specific_info(self, object_info):
        type_specific_inspector = None
        for insp in self.type_specific_inspectors:
            if insp.applies_to(object_info):
                type_specific_inspector = insp
                break
        
        if type_specific_inspector != self.current_type_specific_inspector:
            if self.current_type_specific_inspector is not None:
                self.current_type_specific_inspector.grid_remove() # TODO: or forget?
                self.current_type_specific_label.destroy()
                self.current_type_specific_inspector = None
                self.current_type_specific_label = None
                
            if type_specific_inspector is not None:
                self.current_type_specific_label = self._add_block_label (3, "")
                
                type_specific_inspector.grid(row=4, 
                                             column=0, 
                                             columnspan=4, 
                                             sticky=tk.NSEW,
                                             padx=(0,10))
                
            self.current_type_specific_inspector = type_specific_inspector
        
        if self.current_type_specific_inspector is not None:
            self.current_type_specific_inspector.set_object_info(object_info,
                                                             self.current_type_specific_label)
    
    def goto_type(self, event):
        if self.object_info is not None:
            get_workbench().event_generate("ObjectSelect", object_id=self.object_info["type_id"])
    
    
    
    def _add_block_label(self, row, caption):
        label = tk.Label(self.grid_frame, bg=CALM_WHITE, text=caption)
        label.grid(row=row, column=0, columnspan=4, sticky="nsew", pady=(20,0))
        return label
            

class TypeSpecificInspector:
    def __init__(self, master):
        pass
    
    def set_object_info(self, object_info, label):
        pass
    
    def applies_to(self, object_info):
        return False
    
class FileHandleInspector(TextFrame, TypeSpecificInspector):
    
    def __init__(self, master):
        TypeSpecificInspector.__init__(self, master)
        TextFrame.__init__(self, master, read_only=True)
        self.cache = {} # stores file contents for handle id-s
        self.config(borderwidth=1)
        self.text.configure(background="white")
        self.text.tag_configure("read", foreground="lightgray")
    
    def applies_to(self, object_info):
        return ("file_content" in object_info
                or "file_error" in object_info)
    
    def set_object_info(self, object_info, label):
        
        if "file_content" not in object_info:
            exception("File error: " + object_info["file_error"])
            return
        
        assert "file_content" in object_info
        content = object_info["file_content"]
        line_count_sep = len(content.split("\n"))
        line_count_term = len(content.splitlines())
        char_count = len(content)
        self.text.configure(height=min(line_count_sep, 10))
        self.text.set_content(content)
        
        assert "file_tell" in object_info
        # f.tell() gives num of bytes read (minus some magic with linebreaks)
        
        file_bytes = content.encode(encoding=object_info["file_encoding"])
        bytes_read = file_bytes[0:object_info["file_tell"]]
        read_content = bytes_read.decode(encoding=object_info["file_encoding"])
        read_char_count = len(read_content)
        read_line_count_term = (len(content.splitlines())
                                - len(content[read_char_count:].splitlines()))
        
        pos_index = "1.0+" + str(read_char_count) + "c"
        self.text.tag_add("read", "1.0", pos_index)
        self.text.see(pos_index)
        
        label.configure(text="Read %d/%d %s, %d/%d %s" 
                        % (read_char_count,
                           char_count,
                           "symbol" if char_count == 1 else "symbols",  
                           read_line_count_term,
                           line_count_term,
                           "line" if line_count_term == 1 else "lines"))
            
            
            
class FunctionInspector(TextFrame, TypeSpecificInspector):
    
    def __init__(self, master):
        TypeSpecificInspector.__init__(self, master)
        TextFrame.__init__(self, master, read_only=True)
        self.config(borderwidth=1)
        self.text.configure(background="white")

    def applies_to(self, object_info):
        return "source" in object_info
    
    def set_object_info(self, object_info, label):
        line_count = len(object_info["source"].split("\n"))
        self.text.configure(height=min(line_count, 15))
        self.text.set_content(object_info["source"])
        label.configure(text="Code")
                
            
class StringInspector(TextFrame, TypeSpecificInspector):
    
    def __init__(self, master):
        TypeSpecificInspector.__init__(self, master)
        TextFrame.__init__(self, master, read_only=True)
        self.config(borderwidth=1)
        self.text.configure(background="white")

    def applies_to(self, object_info):
        return object_info["type"] == repr(str)
    
    def set_object_info(self, object_info, label):
        # TODO: don't show too big string
        content = ast.literal_eval(object_info["repr"])
        line_count_sep = len(content.split("\n"))
        line_count_term = len(content.splitlines())
        self.text.configure(height=min(line_count_sep, 10))
        self.text.set_content(content)
        label.configure(text="%d %s, %d %s" 
                        % (len(content),
                           "symbol" if len(content) == 1 else "symbols",
                           line_count_term, 
                           "line" if line_count_term == 1 else "lines"))
        

class ElementsInspector(MemoryFrame, TypeSpecificInspector):
    def __init__(self, master):
        TypeSpecificInspector.__init__(self, master)
        MemoryFrame.__init__(self, master, ('index', 'id', 'value'))
        self.configure(border=1)
        
        #self.vert_scrollbar.grid_remove()
        self.tree.column('index', width=40, anchor=tk.W, stretch=False)
        self.tree.column('id', width=750, anchor=tk.W, stretch=True)
        self.tree.column('value', width=750, anchor=tk.W, stretch=True)
        
        self.tree.heading('index', text='Index', anchor=tk.W) 
        self.tree.heading('id', text='Value ID', anchor=tk.W)
        self.tree.heading('value', text='Value', anchor=tk.W)
    
        self.elements_have_indices = None
        self.update_memory_model()

        get_workbench().bind("ShowView", self.update_memory_model, True)
        get_workbench().bind("HideView", self.update_memory_model, True)
        
        
    def update_memory_model(self, event=None):
        self._update_columns()
        
    def _update_columns(self):
        if get_workbench().in_heap_mode():
            if self.elements_have_indices:
                self.tree.configure(displaycolumns=("index", "id"))
            else:
                self.tree.configure(displaycolumns=("id",))
        else:
            if self.elements_have_indices:
                self.tree.configure(displaycolumns=("index", "value"))
            else:
                self.tree.configure(displaycolumns=("value"))

    def applies_to(self, object_info):
        return "elements" in object_info
    
    def on_select(self, event):
        pass
    
    def on_double_click(self, event):
        self.show_selected_object_info()
    
    def set_object_info(self, object_info, label):
        assert "elements" in object_info
        
        self.elements_have_indices = object_info["type"] in (repr(tuple), repr(list))
        self._update_columns()
        
        self._clear_tree()
        index = 0
        # TODO: don't show too big number of elements
        for element in object_info["elements"]:
            node_id = self.tree.insert("", "end")
            if self.elements_have_indices:
                self.tree.set(node_id, "index", index)
            else:
                self.tree.set(node_id, "index", "")
                
            self.tree.set(node_id, "id", format_object_id(element["id"]))
            self.tree.set(node_id, "value", shorten_repr(element["repr"], MAX_REPR_LENGTH_IN_GRID))
            index += 1

        count = len(object_info["elements"])
        self.tree.config(height=min(count,10))
        
        
        label.configure (
            text=("%d element" if count == 1 else "%d elements") % count
        ) 
        

class DictInspector(MemoryFrame, TypeSpecificInspector):
    def __init__(self, master):
        TypeSpecificInspector.__init__(self, master)
        MemoryFrame.__init__(self, master, ('key_id', 'id', 'key', 'value'))
        self.configure(border=1)
        #self.vert_scrollbar.grid_remove()
        self.tree.column('key_id', width=100, anchor=tk.W, stretch=False)
        self.tree.column('key', width=100, anchor=tk.W, stretch=False)
        self.tree.column('id', width=750, anchor=tk.W, stretch=True)
        self.tree.column('value', width=750, anchor=tk.W, stretch=True)
        
        self.tree.heading('key_id', text='Key ID', anchor=tk.W) 
        self.tree.heading('key', text='Key', anchor=tk.W) 
        self.tree.heading('id', text='Value ID', anchor=tk.W)
        self.tree.heading('value', text='Value', anchor=tk.W)
    
        self.update_memory_model()
        
    def update_memory_model(self, event=None):
        if get_workbench().in_heap_mode():
            self.tree.configure(displaycolumns=("key_id", "id"))
        else:
            self.tree.configure(displaycolumns=("key", "value"))

    def applies_to(self, object_info):
        return "entries" in object_info

    def on_select(self, event):
        pass
    
    def on_double_click(self, event):
        # NB! this selects value
        self.show_selected_object_info()

    def set_object_info(self, object_info, label):
        assert "entries" in object_info
        
        self._clear_tree()
        # TODO: don't show too big number of elements
        for key, value in object_info["entries"]:
            node_id = self.tree.insert("", "end")
            self.tree.set(node_id, "key_id", format_object_id(key["id"]))
            self.tree.set(node_id, "key", shorten_repr(key["repr"], MAX_REPR_LENGTH_IN_GRID))
            self.tree.set(node_id, "id", format_object_id(value["id"]))
            self.tree.set(node_id, "value", shorten_repr(value["repr"], MAX_REPR_LENGTH_IN_GRID))

        count = len(object_info["entries"])
        self.tree.config(height=min(count,10))
        
        label.configure (
            text=("%d entry" if count == 1 else "%d entries") % count
        ) 
        
        self.update_memory_model()

def load_plugin():
    get_workbench().add_view(ObjectInspector, "Object inspector", "se")


def get_module_node(script):
    if hasattr(script, "_get_module_node"):
        return script._get_module_node()
    elif hasattr(script, "_get_module"):
        return script._get_module()
    else:
        return script._parser.module()

def get_version_tuple():
    import jedi
    nums = []
    for item in jedi.__version__.split("."):
        try:
            nums.append(int(item))
        except:
            nums.append(0)
            
    return tuple(nums)


# This is a proxy module which gives frontend the illusion 
# that common lives directly in thonny package
# (this is the case for backend)
from thonny.shared.thonny.common import *  # @UnusedWildImport


# -*- coding: utf-8 -*-
import tkinter as tk
from tkinter import ttk
from thonny.globals import get_workbench
from thonny.ui_utils import select_sequence

def load_plugin():
    def create_edit_command_handler(virtual_event_sequence):
        def handler(event=None):
            widget = get_workbench().focus_get()
            if widget:
                return widget.event_generate(virtual_event_sequence)
        
        return handler
    
    def select_all(event=None):
        # Tk 8.6 has <<SelectAll>> virtual event, but 8.5 doesn't
        widget = get_workbench().focus_get()
        if isinstance(widget, tk.Text):
            widget.tag_remove("sel","1.0","end")
            widget.tag_add("sel","1.0","end")
        elif isinstance(widget, ttk.Entry) or isinstance(widget, tk.Entry):
            widget.select_range(0, tk.END)
        
    
    get_workbench().add_command("undo", "edit", "Undo",
        create_edit_command_handler("<<Undo>>"),
        tester=None, # TODO:
        default_sequence=select_sequence("<Control-z>", "<Command-z>"),
        skip_sequence_binding=True,
        group=10)
    
    get_workbench().add_command("redo", "edit", "Redo",
        create_edit_command_handler("<<Redo>>"),
        tester=None, # TODO:
        default_sequence=select_sequence("<Control-y>", "<Command-y>"),
        skip_sequence_binding=True,
        group=10)
    
    # Ctrl+Shift+Z as alternative shortcut for redo
    get_workbench().bind_class("Text", select_sequence("<Control-Shift-Z>", "<Command-Shift-Z>"),
                               create_edit_command_handler("<<Redo>>"), True)
    
    
    get_workbench().add_command("Cut", "edit", "Cut",
        create_edit_command_handler("<<Cut>>"),
        tester=None, # TODO:
        default_sequence=select_sequence("<Control-x>", "<Command-x>"),
        skip_sequence_binding=True,
        group=20)
    
    get_workbench().add_command("Copy", "edit", "Copy",
        create_edit_command_handler("<<Copy>>"),
        tester=None, # TODO:
        default_sequence=select_sequence("<Control-c>", "<Command-c>"),
        skip_sequence_binding=True,
        group=20)
    
    get_workbench().add_command("Paste", "edit", "Paste",
        create_edit_command_handler("<<Paste>>"),
        tester=None, # TODO:
        default_sequence=select_sequence("<Control-v>", "<Command-v>"),
        skip_sequence_binding=True,
        group=20)
    
    get_workbench().add_command("SelectAll", "edit", "Select all",
        select_all,
        tester=None, # TODO:
        default_sequence=select_sequence("<Control-a>", "<Command-a>"),
        skip_sequence_binding=True,
        group=20)
    



# -*- coding: utf-8 -*-

_worbench = None
_runner = None

def register_workbench(workbench):
    global _workbench
    _workbench = workbench

def register_runner(runner):
    global _runner
    _runner = runner


def get_workbench():    
    return _workbench

def get_runner():    
    return _runner
    


import os.path
import sys
import runpy

try:
    runpy.run_module("thonny.customize", run_name="__main__")
except ImportError:
    pass


THONNY_USER_DIR = os.environ.get("THONNY_USER_DIR", 
                                 os.path.expanduser(os.path.join("~", ".thonny")))

THONNY_USER_BASE = os.path.join(THONNY_USER_DIR, "plugins")

def launch():
    _prepare_thonny_user_dir()
    
    try:
        _update_sys_path()
        
        from thonny import workbench
        
        if _should_delegate():
            # First check if there is existing Thonny instance to handle the request
            delegation_result = _try_delegate_to_existing_instance(sys.argv[1:])
            if delegation_result == True:
                # we're done
                print("Delegated to an existing Thonny instance. Exiting now.")
                return
            
            if hasattr(delegation_result, "accept"):
                # we have server socket to put in use
                server_socket = delegation_result
            else:
                server_socket = None
                 
            bench = workbench.Workbench(server_socket)
        else:
            bench = workbench.Workbench()
            
        try:
            bench.mainloop()
        except SystemExit:
            bench.destroy()
        return 0
    except SystemExit as e:
        from tkinter import messagebox
        messagebox.showerror("System exit", str(e))
    except:
        from logging import exception
        exception("Internal error")
        import tkinter.messagebox
        import traceback
        tkinter.messagebox.showerror("Internal error", traceback.format_exc())
        return -1
    finally:
        from thonny.globals import get_runner
        runner = get_runner()
        if runner != None:
            runner.kill_backend()


def _prepare_thonny_user_dir():
    if not os.path.exists(THONNY_USER_DIR):
        os.makedirs(THONNY_USER_DIR, mode=0o700, exist_ok=True)
        
        # user_dir_template is a post-installation means for providing
        # alternative default user environment in multi-user setups
        template_dir = os.path.join(os.path.dirname(__file__), "user_dir_template")
                    
        if os.path.isdir(template_dir):
            import shutil
            
            def copy_contents(src_dir, dest_dir):
                # I want the copy to have current user permissions
                for name in os.listdir(src_dir):
                    src_item = os.path.join(src_dir, name)
                    dest_item = os.path.join(dest_dir, name)
                    if os.path.isdir(src_item):
                        os.makedirs(dest_item, mode=0o700)
                        copy_contents(src_item, dest_item)
                    else:
                        shutil.copyfile(src_item, dest_item)
                        os.chmod(dest_item, 0o600)
                        
            copy_contents(template_dir, THONNY_USER_DIR)
            

def _update_sys_path():
    import site
    
    # remove old dir from path
    if site.getusersitepackages() in sys.path:
        sys.path.remove(site.getusersitepackages())
        
    # compute usersitepackages that plugins installation subprocess would see
    import subprocess
    env = os.environ.copy()
    env["PYTHONUSERBASE"] = THONNY_USER_BASE
    proc = subprocess.Popen(
        [sys.executable.replace("thonny.exe", "pythonw.exe"),
         "-c", "import site; print(site.getusersitepackages())"],
        universal_newlines=True, env=env, stdout=subprocess.PIPE)
    plugins_sitepackages = proc.stdout.readline().strip()
    
    sys.path.append(plugins_sitepackages)

def _should_delegate():
    from thonny import workbench
    from thonny.config import try_load_configuration
    configuration_manager = try_load_configuration(workbench.CONFIGURATION_FILE_NAME)
    # Setting the default
    configuration_manager.set_default("general.single_instance", workbench.SINGLE_INSTANCE_DEFAULT)
    # getting the value (may use the default or return saved value)
    return configuration_manager.get_option("general.single_instance")

def _try_delegate_to_existing_instance(args):
    import socket
    from thonny import workbench
    try:
        # Try to create server socket.
        # This is fastest way to find out if Thonny is already running
        serversocket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        serversocket.bind(("localhost", workbench.THONNY_PORT))
        serversocket.listen(10)
        # we were able to create server socket (ie. Thonny was not running)
        # Let's use the socket in Thonny so that requests coming while 
        # UI gets constructed don't get lost.
        # (Opening several files with Thonny in Windows results in many
        # Thonny processes opened quickly) 
        return serversocket
    except OSError:
        # port was already taken, most likely by previous Thonny instance.
        # Try to connect and send arguments
        try:
            return _delegate_to_existing_instance(args)
        except:
            import traceback
            traceback.print_exc()
            return False
        
        
def _delegate_to_existing_instance(args):
    import socket
    from thonny import workbench
    data = repr(args).encode(encoding='utf_8')
    sock = socket.create_connection(("localhost", workbench.THONNY_PORT))
    sock.sendall(data)
    sock.shutdown(socket.SHUT_WR)
    response = bytes([])
    while len(response) < len(workbench.SERVER_SUCCESS):
        new_data = sock.recv(2)
        if len(new_data) == 0:
            break
        else:
            response += new_data
    
    return response.decode("UTF-8") == workbench.SERVER_SUCCESS


    
def get_version():
    try:
        package_dir = os.path.dirname(sys.modules["thonny"].__file__)
        with open(os.path.join(package_dir, "VERSION"), encoding="ASCII") as fp:
            return fp.read().strip()
    except:
        return "0.0.0"
      
    



"""
from rope.base.project import Project
from rope.refactor.rename import Rename
import rope.base.libutils
import tkinter as tk
from tkinter import ttk
import os.path

#arguments: 1) full path to the file, 2) new name to be applied to the identifier, 3) Rope offset (position) of the renamed identifier in the current file
#returns a list of Rope change objects applying to this rename refactor
#throws an exception if anything goes wrong, needs to be handled by callers!
def get_list_of_rename_changes(full_path, new_variable_name, offset):
    filearr = os.path.split(full_path)
    project_path = filearr[0]
    module_name = filearr[1]
    project = Project(project_path, ropefolder=None)
    module = rope.base.libutils.path_to_resource(project, full_path)
    changes = Rename(project, module, offset).get_changes(new_variable_name)
    return (project, changes)
    
#performs the changes 
#arguments: 1) Rope project, 2) Rope changes object
def perform_changes(project, changes):
    project.do(changes)
    project.close()

#cancels the changes, cleans up the project
def cancel_changes(project):
    project.close()

#utility method for convering a Text index to a Rope offset
def calculate_offset(text):
    contents = text.get(1.0, 'end').split('\n')
    insert_index = text.index('insert')
    linearr = insert_index.split('.')
    line_no = int(linearr[0])
    char_no = int(linearr[1])

    totalchars = char_no
    for line in range(line_no - 1):
        totalchars += len(contents[line]) + 1

    return totalchars

#inspired by http://effbot.org/tkinterbook/tkinter-dialog-windows.htm
#creates a window asking for a new identifier name, can later be accessed via the refactor_new_variable_name variable
class RenameWindow(tk.Toplevel):
    def __init__(self, master, title = None):
        tk.Toplevel.__init__(self, master)
        self.refactor_new_variable_name = None

        self.transient(master)

        self.title('Rename')

        self.parent = master
        self.result = None
    
        ttk.Label(self, text="New name:").grid(row=0, columnspan=2)
        self.new_name_entry = ttk.Entry(self)
        self.new_name_entry.grid(row=1, columnspan=2)
        self.new_name_entry.focus_force();

        self.ok_button = ttk.Button(self, text="OK", command=self.ok, default=tk.ACTIVE)
        self.cancel_button = ttk.Button(self, text="Cancel", command=self.cancel)
        self.ok_button.grid(row=2, column=0, sticky=tk.W + tk.E, padx=5)
        self.cancel_button.grid(row=2, column=1, sticky=tk.W + tk.E, padx=5)

        self.bind("<Return>", self.ok, True)
        self.bind("<Escape>", self.cancel, True)

        self.grab_set()

        self.protocol("WM_DELETE_WINDOW", self.cancel)
        self.geometry("+%d+%d" % (master.winfo_rootx()+50,
                                  master.winfo_rooty()+50))
        self.resizable(width=False, height=False)

        self.wait_window(self)
    
    #user clicked cancel - destroy the object, return the focus to master    
    def cancel(self, event=None):
        self.parent.focus_set()
        self.destroy()

    #user clicked ok - set the variable name, destroy the object, return the focus to parent
    def ok(self, event=None):
        self.withdraw()
        self.update_idletasks()
        self.refactor_new_variable_name = self.new_name_entry.get()
        self.cancel()


class RefactorRenameStartEvent(thonny.user_logging.UserEvent): #user initiated the refactoring process
    def __init__(self, editor):
        self.editor_id = id(editor)

class RefactorRenameCancelEvent(thonny.user_logging.UserEvent): #user manually cancelled the refactoring process
    def __init__(self, editor):
        self.editor_id = id(editor)

class RefactorRenameFailedEvent(thonny.user_logging.UserEvent): #refactoring process failed due to an error
    def __init__(self, editor):
        self.editor_id = id(editor)

class RefactorRenameCompleteEvent(thonny.user_logging.UserEvent): #refactoring process was successfully completed
    def __init__(self, description, offset, affected_files):
        self.description = description
        self.offset = offset
        self.affected_files = affected_files

def _cmd_refactor_rename(self):
    self.log_user_event(thonny.refactor.RefactorRenameStartEvent(self.editor_notebook.get_current_editor()))
    if not self.editor_notebook.get_current_editor():
        self.log_user_event(thonny.refactor.RefactorRenameFailedEvent(self.editor_notebook.get_current_editor()))
        errorMessage = tkMessageBox.showerror(
                       title="Rename failed",
                       message="Rename operation failed (no active editor tabs?).", #TODO - more informative text needed
                       master=self)
        return

    #create a list of active but unsaved/modified editors)
    unsaved_editors = [x for x in self.editor_notebook.winfo_children() if type(x) == Editor and x._cmd_save_file_enabled()]

    if len(unsaved_editors) != 0:
        #confirm with the user that all open editors need to be saved first
        confirm = tkMessageBox.askyesno(
                  title="Save Files Before Rename",
                  message="All modified files need to be saved before refactoring. Do you want to continue?",
                  default=tkMessageBox.YES,
                  master=self)

        if not confirm:
            self.log_user_event(thonny.refactor.RefactorRenameCancelEvent(self.editor_notebook.get_current_editor()))
            return #if user doesn't want it, return

        for editor in unsaved_editors:                     
            if not editor.get_filename():
                self.editor_notebook.select(editor) #in the case of editors with no filename, show it, so user knows which one they're saving
            editor._cmd_save_file()
            if editor._cmd_save_file_enabled(): #just a sanity check - if after saving a file still needs saving, something is wrong
                self.log_user_event(thonny.refactor.RefactorRenameFailedEvent(self.editor_notebook.get_current_editor()))
                errorMessage = tkMessageBox.showerror(
                               title="Rename failed",
                               message="Rename operation failed (saving file failed).", #TODO - more informative text needed
                               master=self)
                return

    filename = self.editor_notebook.get_current_editor().get_filename()

    if filename == None: #another sanity check - the current editor should have an associated filename by this point 
        self.log_user_event(thonny.refactor.RefactorRenameFailedEvent(self.editor_notebook.get_current_editor()))
        errorMessage = tkMessageBox.showerror(
                       title="Rename failed",
                       message="Rename operation failed (no filename associated with current module).", #TODO - more informative text needed
                       master=self)
        return

    identifier = re.compile(r"^[^\d\W]\w*\Z", re.UNICODE) #regex to compare valid python identifiers against

    while True: #ask for new variable name until a valid one is entered
        renameWindow = thonny.refactor.RenameWindow(self)
        newname = renameWindow.refactor_new_variable_name
        if newname == None:
            self.log_user_event(thonny.refactor.RefactorRenameCancelEvent(self.editor_notebook.get_current_editor()))
            return #user canceled, return

        if re.match(identifier, newname):
            break #valid identifier entered, continue

        errorMessage = tkMessageBox.showerror(
                       title="Incorrect identifier",
                       message="Incorrect Python identifier, please re-enter.",
                       master=self)

    try: 
        #calculate the offset for rope
        offset = thonny.refactor.calculate_offset(self.editor_notebook.get_current_editor()._code_view.text)
        #get the project handle and list of changes
        project, changes = thonny.refactor.get_list_of_rename_changes(filename, newname, offset)
        #if len(changes.changes == 0): raise Exception

    except Exception:
        try: #rope needs the cursor to be AFTER the first character of the variable being refactored
             #so the reason for failure might be that the user had the cursor before the variable name
            offset = offset + 1
            project, changes = thonny.refactor.get_list_of_rename_changes(filename, newname, offset)
            #if len(changes.changes == 0): raise Exception

        except Exception: #couple of different reasons why this could happen, let's list them all in the error message
            self.log_user_event(thonny.refactor.RefactorRenameFailedEvent(self.editor_notebook.get_current_editor()))
            message = 'Rename operation failed. A few possible reasons: \n'
            message += '1) Not a valid Python identifier selected \n'
            message += '2) The current file or any other files in the same directory or in any of its subdirectores contain incorrect syntax. Make sure the current project is in its own separate folder.'
            errorMessage = tkMessageBox.showerror(
                           title="Rename failed",
                           message=message, #TODO - maybe also print stack trace for more info?
                           master=self)               
            return

    description = changes.description #needed for logging

    #sanity check
    if len(changes.changes) == 0:
        self.log_user_event(thonny.refactor.RefactorRenameFailedEvent(self.editor_notebook.get_current_editor()))
        errorMessage = tkMessageBox.showerror(
                           title="Rename failed",
                           message="Rename operation failed - no identifiers affected by change.", #TODO - more informative text needed
                           master=self)               
        return

    affected_files = [] #needed for logging
    #show the preview window to user
    messageText = 'Confirm the changes. The following files will be modified:\n'
    for change in changes.changes:
        affected_files.append(change.resource._path)
        messageText += '\n ' + change.resource._path

    messageText += '\n\n NB! This action cannot be undone.'

    confirm = tkMessageBox.askyesno(
              title="Confirm changes",
              message=messageText,
              default=tkMessageBox.YES,
              master=self)
    
    #confirm with user to finalize the changes
    if not confirm:
        self.log_user_event(thonny.refactor.RefactorRenameCancelEvent(self.editor_notebook.get_current_editor()))
        thonny.refactor.cancel_changes(project)
        return

    try:
        thonny.refactor.perform_changes(project, changes)
    except Exception:
            self.log_user_event(thonny.refactor.RefactorRenameFailedEvent(self.editor_notebook.get_current_editor()))
            errorMessage = tkMessageBox.showerror(
                           title="Rename failed",
                           message="Rename operation failed (Rope error).", #TODO - more informative text needed
                           master=self)     
            thonny.refactor.cancel_changes(project)
            return            

    #everything went fine, let's load all the active tabs again and set their content
    for editor in self.editor_notebook.winfo_children():
        try: 
            filename = editor.get_filename()
            source, self.file_encoding = misc_utils.read_python_file(filename)
            editor._code_view.set_content(source)
            self.editor_notebook.tab(editor, text=self.editor_notebook._generate_editor_title(filename))
        except Exception:
            try: #it is possible that a file (module) itself was renamed - Rope allows it. so let's see if a file exists with the new name. 
                filename = filename.replace(os.path.split(filename)[1], newname + '.py')
                source, self.file_encoding = misc_utils.read_python_file(filename)
                editor._code_view.set_content(source)
                self.editor_notebook.tab(editor, text=self.editor_notebook._generate_editor_title(filename))
            except Exception: #something went wrong with reloading the file, let's close this tab to avoid consistency problems
                self.editor_notebook.forget(editor)
                editor.destroy()

    self.log_user_event(thonny.refactor.RefactorRenameCompleteEvent(description, offset, affected_files))
    current_browser_node_path = self.file_browser.get_selected_path()
    self.file_browser.refresh_tree()
    if current_browser_node_path is not None:
        self.file_browser.open_path_in_browser(current_browser_node_path)

def _cmd_refactor_rename_enabled(self):
    return self.editor_notebook.get_current_editor() is not None

def _load_plugin_(workbench):
    get_workbench().add_command("refactor_rename", "edit", "Rename identifier", ...)
"""


import tkinter as tk
from tkinter import ttk
from thonny.globals import get_workbench


class ConfigurationDialog(tk.Toplevel):
    def __init__(self, master, page_records):
        tk.Toplevel.__init__(self, master)
        width = 400
        height = 400
        left = max(int(get_workbench().winfo_x() + get_workbench().winfo_width()/2 - width/2), 0)
        top = max(int(get_workbench().winfo_y() + get_workbench().winfo_height()/2 - height/2), 0)
        self.geometry("%dx%d+%d+%d" % (width, height, left, top))
        self.title("Thonny options")
        
        self.columnconfigure(0, weight=1)
        self.rowconfigure(0, weight=1)
        
        main_frame = ttk.Frame(self) # otherwise there is wrong color background with clam
        main_frame.grid(row=0, column=0, sticky=tk.NSEW)
        main_frame.columnconfigure(0, weight=1)
        main_frame.rowconfigure(0, weight=1)
        
        self._notebook = ttk.Notebook(main_frame)
        self._notebook.grid(row=0, column=0, columnspan=3, sticky=tk.NSEW, padx=10, pady=10)
        
        self._ok_button = ttk.Button(main_frame, text="OK", command=self._ok, default="active")
        self._cancel_button = ttk.Button(main_frame, text="Cancel", command=self._cancel)
        self._ok_button.grid(row=1, column=1, padx=(0,11), pady=(0,10))
        self._cancel_button.grid(row=1, column=2, padx=(0,11), pady=(0,10))
        
        self._pages = {}
        for title in sorted(page_records):
            page_class = page_records[title]
            spacer = ttk.Frame(self)
            spacer.rowconfigure(0, weight=1)
            spacer.columnconfigure(0, weight=1)
            page = page_class(spacer)
            self._pages[title] = page
            page.grid(sticky=tk.NSEW, pady=10, padx=15)
            self._notebook.add(spacer, text=title)
        
        self.bind("<Return>", self._ok, True)
        self.bind("<Escape>", self._cancel, True)
    
    def _ok(self, event=None):
        for title in sorted(self._pages):
            try:
                page = self._pages[title]
                if page.apply() == False:
                    return
            except:
                get_workbench().report_exception("Error when applying options in " + title)
             
        self.destroy()
    
    def _cancel(self, event=None):
        self.destroy()

class ConfigurationPage(ttk.Frame):
    """This is an example dummy implementation of a configuration page.
    
    It's not required that configuration pages inherit from this class
    (can be any widget), but the class must have constructor with single parameter
    for getting the master."""
    def __init__(self, master):
        ttk.Frame.__init__(self, master)
    
    def add_checkbox(self, flag_name, description, row=None, pady=0, columnspan=1):
        variable = get_workbench().get_variable(flag_name)
        checkbox = ttk.Checkbutton(self, text=description, variable=variable)
        checkbox.grid(row=row, column=0, sticky=tk.W, pady=pady, columnspan=columnspan)
                
    def apply(self):
        """Apply method should return False, when page contains invalid
        input and configuration dialog should not be closed."""
        pass
        


# -*- coding: utf-8 -*-

import webbrowser

import tkinter as tk
from tkinter import ttk, messagebox

from thonny import misc_utils, tktextext, ui_utils, THONNY_USER_BASE
from thonny.globals import get_workbench, get_runner
import subprocess
from urllib.request import urlopen, urlretrieve
import urllib.error
import urllib.parse
from concurrent.futures.thread import ThreadPoolExecutor
import os
import json
from distutils.version import LooseVersion, StrictVersion
import logging
import re
from tkinter.filedialog import askopenfilename
from logging import exception
from thonny.ui_utils import SubprocessDialog, AutoScrollbar, get_busy_cursor
from thonny.misc_utils import running_on_windows
import sys

LINK_COLOR="#3A66DD"
PIP_INSTALLER_URL="https://bootstrap.pypa.io/get-pip.py"

class PipDialog(tk.Toplevel):
    def __init__(self, master, only_user=False):
        self._state = None # possible values: "listing", "fetching", "idle"
        self._process = None
        self._installed_versions = {}
        self._only_user = only_user
        self.current_package_data = None
        
        tk.Toplevel.__init__(self, master)
        
        main_frame = ttk.Frame(self)
        main_frame.grid(sticky=tk.NSEW, ipadx=15, ipady=15)
        self.rowconfigure(0, weight=1)
        self.columnconfigure(0, weight=1)

        self.title(self._get_title())
        if misc_utils.running_on_mac_os():
            self.configure(background="systemSheetBackground")
        self.transient(master)
        self.grab_set() # to make it active
        #self.grab_release() # to allow eg. copy something from the editor 
        
        self._create_widgets(main_frame)
        
        self.search_box.focus_set()
        
        self.bind('<Escape>', self._on_close, True) 
        self.protocol("WM_DELETE_WINDOW", self._on_close)
        self._show_instructions()
        ui_utils.center_window(self, master)
        
        self._start_update_list()
    
    
    def _create_widgets(self, parent):
        
        header_frame = ttk.Frame(parent)
        header_frame.grid(row=1, column=0, sticky="nsew", padx=15, pady=(15,0))
        header_frame.columnconfigure(0, weight=1)
        header_frame.rowconfigure(1, weight=1)
        
        name_font = tk.font.nametofont("TkDefaultFont").copy()
        name_font.configure(size=16)
        self.search_box = ttk.Entry(header_frame, background=ui_utils.CALM_WHITE)
        self.search_box.grid(row=1, column=0, sticky="nsew")
        self.search_box.bind("<Return>", self._on_search, False)
        
        self.search_button = ttk.Button(header_frame, text="Search", command=self._on_search)
        self.search_button.grid(row=1, column=1, sticky="nse", padx=(10,0))
        
        
        main_pw = tk.PanedWindow(parent, orient=tk.HORIZONTAL,
                                 background=ui_utils.get_main_background(),
                                 sashwidth=10)
        main_pw.grid(row=2, column=0, sticky="nsew", padx=15, pady=15)
        parent.rowconfigure(2, weight=1)
        parent.columnconfigure(0, weight=1)
        
        listframe = ttk.Frame(main_pw, relief="groove", borderwidth=1)
        listframe.rowconfigure(0, weight=1)
        listframe.columnconfigure(0, weight=1)
        
        self.listbox = tk.Listbox(listframe, activestyle="dotbox", 
                                  width=20, height=10,
                                  background=ui_utils.CALM_WHITE,
                                  selectborderwidth=0, relief="flat",
                                  highlightthickness=0, borderwidth=0)
        self.listbox.insert("end", " <INSTALL>")
        self.listbox.bind("<<ListboxSelect>>", self._on_listbox_select, True)
        self.listbox.grid(row=0, column=0, sticky="nsew")
        list_scrollbar = AutoScrollbar(listframe, orient=tk.VERTICAL)
        list_scrollbar.grid(row=0, column=1, sticky="ns")
        list_scrollbar['command'] = self.listbox.yview
        self.listbox["yscrollcommand"] = list_scrollbar.set
        
        info_frame = ttk.Frame(main_pw)
        info_frame.columnconfigure(0, weight=1)
        info_frame.rowconfigure(1, weight=1)
        
        main_pw.add(listframe)
        main_pw.add(info_frame)
        
        self.name_label = ttk.Label(info_frame, text="", font=name_font)
        self.name_label.grid(row=0, column=0, sticky="w", padx=5)
        

        
        info_text_frame = tktextext.TextFrame(info_frame, read_only=True,
                                              horizontal_scrollbar=False,
                                              vertical_scrollbar_class=AutoScrollbar,
                                              width=60, height=10)
        info_text_frame.configure(borderwidth=1)
        info_text_frame.grid(row=1, column=0, columnspan=4, sticky="nsew", pady=(0,20))
        self.info_text = info_text_frame.text
        self.info_text.tag_configure("url", foreground=LINK_COLOR, underline=True)
        self.info_text.tag_bind("url", "<ButtonRelease-1>", self._handle_url_click)
        self.info_text.tag_bind("url", "<Enter>", lambda e: self.info_text.config(cursor="hand2"))
        self.info_text.tag_bind("url", "<Leave>", lambda e: self.info_text.config(cursor=""))
        self.info_text.tag_configure("install_file", foreground=LINK_COLOR, underline=True)
        self.info_text.tag_bind("install_file", "<ButtonRelease-1>", self._handle_install_file_click)
        self.info_text.tag_bind("install_file", "<Enter>", lambda e: self.info_text.config(cursor="hand2"))
        self.info_text.tag_bind("install_file", "<Leave>", lambda e: self.info_text.config(cursor=""))
        
        default_font = tk.font.nametofont("TkDefaultFont")
        self.info_text.configure(background=ui_utils.get_main_background(),
                                 font=default_font,
                                 wrap="word")

        bold_font = default_font.copy()
        # need to explicitly copy size, because Tk 8.6 on certain Ubuntus use bigger font in copies
        bold_font.configure(weight="bold", size=default_font.cget("size"))
        self.info_text.tag_configure("caption", font=bold_font)
        
        
        self.command_frame = ttk.Frame(info_frame)
        self.command_frame.grid(row=2, column=0, sticky="w")
        
        self.install_button = ttk.Button(self.command_frame, text=" Upgrade ",
                                         command=self._on_click_install)
        
        if not self._read_only():
            self.install_button.grid(row=0, column=0, sticky="w", padx=0)
        
        self.uninstall_button = ttk.Button(self.command_frame, text="Uninstall",
                                           command=lambda: self._perform_action("uninstall"))
        
        if not self._read_only():
            self.uninstall_button.grid(row=0, column=1, sticky="w", padx=(5,0))
        
        self.advanced_button = ttk.Button(self.command_frame, text="...", width=3,
                                          command=lambda: self._perform_action("advanced"))
        
        if not self._read_only():
            self.advanced_button.grid(row=0, column=2, sticky="w", padx=(5,0))
        
        self.close_button = ttk.Button(info_frame, text="Close", command=self._on_close)
        self.close_button.grid(row=2, column=3, sticky="e")
    
    def _on_click_install(self):
        name = self.current_package_data["info"]["name"]
        if self._confirm_install(name):
            self._perform_action("install")

    def _set_state(self, state, normal_cursor=False):
        self._state = state
        widgets = [self.listbox, 
                           # self.search_box, # looks funny when disabled 
                           self.search_button,
                           self.install_button, self.advanced_button, self.uninstall_button]
        
        if state == "idle":
            self.config(cursor="")
            for widget in widgets:
                widget["state"] = tk.NORMAL
        else:
            self.config(cursor=get_busy_cursor())
            for widget in widgets:
                widget["state"] = tk.DISABLED
        
        if normal_cursor:
            self.config(cursor="")
    
    def _get_state(self):
        return self._state
    
    def _handle_outdated_or_missing_pip(self):
        raise NotImplementedError()
        
    def _install_pip(self):
        self._clear()
        self.info_text.direct_insert("end", "Installing pip\n\n", ("caption", ))
        self.info_text.direct_insert("end", "pip, a required module for managing packages is missing or too old.\n\n"
                                + "Downloading pip installer (about 1.5 MB), please wait ...\n")
        self.update()
        self.update_idletasks()
        
        installer_filename, _ = urlretrieve(PIP_INSTALLER_URL)
        
        self.info_text.direct_insert("end", "Installing pip, please wait ...\n")
        self.update()
        self.update_idletasks()
        
        proc, _ = self._create_backend_process([installer_filename], stderr=subprocess.PIPE)
        out, err = proc.communicate()
        os.remove(installer_filename)
        
        if err != "":
            raise RuntimeError("Error while installing pip:\n" + err)
        
        self.info_text.direct_insert("end", out  + "\n")
        self.update()
        self.update_idletasks()
        
        # update list
        self._start_update_list()
        
        
    def _provide_pip_install_instructions(self):
        self._clear()
        self.info_text.direct_insert("end", "Outdated or missing pip\n\n", ("caption", ))
        self.info_text.direct_insert("end", "pip, a required module for managing packages is missing or too old for Thonny.\n\n"
                                + "If your system package manager doesn't provide recent pip (9.0.0 or later), "
                                + "then you can install newest version by downloading ")
        self.info_text.direct_insert("end", PIP_INSTALLER_URL, ("url",))
        self.info_text.direct_insert("end", " and running it with " 
                                     + self._get_interpreter()
                                     + " (probably needs admin privileges).\n\n")
        
        self.info_text.direct_insert("end", self._instructions_for_command_line_install())
        self._set_state("disabled", True)
    
    def _instructions_for_command_line_install(self):
        return ("Alternatively, if you have an older pip installed, then you can install packages "
                                     + "on the command line (Tools → Open system shell...)")
    
    def _start_update_list(self, name_to_show=None):
        assert self._get_state() in [None, "idle"]
        self._set_state("listing")
        args = ["list"]
        if self._only_user:
            args.append("--user")
        args.extend(["--pre", "--format", "json"])
        self._process, _ = self._create_pip_process(args)
        
        def poll_completion():
            if self._process == None:
                return
            else:
                returncode = self._process.poll()
                if returncode is None:
                    # not done yet
                    self.after(200, poll_completion)
                else:
                    self._set_state("idle")
                    if returncode == 0:
                        raw_data = self._process.stdout.read()
                        self._update_list(json.loads(raw_data))
                        if name_to_show is None:
                            self._show_instructions()
                        else:
                            self._start_show_package_info(name_to_show)
                    else:   
                        error = self._process.stdout.read()
                        if ("no module named pip" in error.lower() # pip not installed
                            or "no such option" in error.lower()): # too old pip
                            self._handle_outdated_or_missing_pip()
                            return
                        else:
                            messagebox.showerror("pip list error", error)
                    
                    self._process = None
        
        poll_completion()
    
    def _update_list(self, data):
        self.listbox.delete(1, "end")
        self._installed_versions = {entry["name"] : entry["version"] for entry in data}
        for name in sorted(self._installed_versions.keys(), key=str.lower):
            self.listbox.insert("end", " " + name)
        
        
    
    def _on_listbox_select(self, event):
        selection = self.listbox.curselection()
        if len(selection) == 1:
            if selection[0] == 0: # special first item
                self._show_instructions()
            else:
                self._start_show_package_info(self.listbox.get(selection[0]).strip())
    
    def _on_search(self, event=None):
        if not self._get_state() == "idle":
            # Search box is not made inactive for busy-states
            return
        
        if self.search_box.get().strip() == "":
            return
        
        self._start_show_package_info(self.search_box.get().strip())
    
    def _clear(self):
        self.current_package_data = None
        self.name_label.grid_remove()
        self.command_frame.grid_remove()
        self.info_text.direct_delete("1.0", "end")
    
    def _show_instructions(self):
        self._clear()
        if self._read_only():
            self.info_text.direct_insert("end", "With current interpreter you can only browse the packages here.\n"
                                       + "Use 'Tools → Open system shell...' for installing, upgrading or uninstalling.")
        else:            
            self.info_text.direct_insert("end", "Install from PyPI\n", ("caption",))
            self.info_text.direct_insert("end", "If you don't know where to get the package from, "
                                         + "then most likely you'll want to search the Python Package Index. "
                                         + "Start by entering the name of the package in the search box above and pressing ENTER.\n\n")
            
            self.info_text.direct_insert("end", "Install from local file\n", ("caption",))
            self.info_text.direct_insert("end", "Click ")
            self.info_text.direct_insert("end", "here", ("install_file",))
            self.info_text.direct_insert("end", " to locate and install the package file (usually with .whl, .tar.gz or .zip extension).\n\n")
            
            self.info_text.direct_insert("end", "Upgrade or uninstall\n", ("caption",))
            self.info_text.direct_insert("end", 'Start by selecting the package from the left.')
        self._select_list_item(0)
    
    def _start_show_package_info(self, name):
        self.current_package_data = None
        self.info_text.direct_delete("1.0", "end")
        self.name_label["text"] = ""
        self.name_label.grid()
        self.command_frame.grid()
        
        installed_version = self._get_installed_version(name) 
        if installed_version is not None:
            self.name_label["text"] = name
            self.info_text.direct_insert("end", "Installed version: ", ('caption',))
            self.info_text.direct_insert("end", installed_version + "\n")
        
        
        # Fetch info from PyPI  
        self._set_state("fetching")
        # Follwing url fetches info about latest version.
        # This is OK even when we're looking an installed older version
        # because new version may have more relevant and complete info.
        url = "https://pypi.python.org/pypi/{}/json".format(urllib.parse.quote(name))
        url_future = _fetch_url_future(url)
            
        def poll_fetch_complete():
            if url_future.done():
                self._set_state("idle")
                try:
                    _, bin_data = url_future.result()
                    raw_data = bin_data.decode("UTF-8")
                    self._show_package_info(name, json.loads(raw_data))
                except urllib.error.HTTPError as e:
                    self._show_package_info(name, self._generate_minimal_data(name), e.code)
                        
            else:
                self.after(200, poll_fetch_complete)
        
        poll_fetch_complete()
    
    def _generate_minimal_data(self, name):
        return {
            "info" : {'name' : name},
            "releases" : {}
            }

    def _show_package_info(self, name, data, error_code=None):
        self.current_package_data = data
        
        def write(s, tag=None):
            if tag is None:
                tags = ()
            else:
                tags = (tag,)
            self.info_text.direct_insert("end", s, tags)
        
        def write_att(caption, value, value_tag=None):
            write(caption + ": ", "caption")
            write(value, value_tag)
            write("\n")
            
        if error_code is not None:
            if error_code == 404:
                write("Could not find the package from PyPI.")
                if not self._get_installed_version(name):
                    # new package
                    write("\nPlease check your spelling!")
                    
            else:
                write("Could not find the package info from PyPI. Error code: " + str(error_code))
            return
        
        info = data["info"]
        self.name_label["text"] = info["name"] # search name could have been a bit different
        latest_stable_version = _get_latest_stable_version(data["releases"].keys())
        if latest_stable_version is not None:
            write_att("Latest stable version", latest_stable_version)
        else:
            write_att("Latest version", data["info"]["version"])
        write_att("Summary", info["summary"])
        write_att("Author", info["author"])
        write_att("Homepage", info["home_page"], "url")
        if info.get("bugtrack_url", None):
            write_att("Bugtracker", info["bugtrack_url"], "url")
        if info.get("docs_url", None):
            write_att("Documentation", info["docs_url"], "url")
        if info.get("package_url", None):
            write_att("PyPI page", info["package_url"], "url")
        
        if self._get_installed_version(info["name"]) is not None:
            self.install_button["text"] = "Upgrade"
            if not self._read_only():
                self.uninstall_button.grid(row=0, column=1)
            
            self._select_list_item(info["name"])
            if self._get_installed_version(info["name"]) == latest_stable_version:
                self.install_button["state"] = "disabled"
            else: 
                self.install_button["state"] = "normal"
        else:
            self.install_button["text"] = "Install"
            self.uninstall_button.grid_forget()
            self._select_list_item(0)
            
    
    def _normalize_name(self, name):
        # looks like (in some cases?) pip list gives the name as it was used during install
        # ie. the list may contain lowercase entry, when actual metadata has uppercase name 
        # Example: when you "pip install cx-freeze", then "pip list"
        # really returns "cx-freeze" although correct name is "cx_Freeze"
        
        # https://www.python.org/dev/peps/pep-0503/#id4
        return re.sub(r"[-_.]+", "-", name).lower().strip()
    
    def _select_list_item(self, name_or_index):
        if isinstance(name_or_index, int):
            index = name_or_index
        else:
            normalized_items = list(map(self._normalize_name, self.listbox.get(0, "end")))
            try:
                index = normalized_items.index(self._normalize_name(name_or_index))
            except:
                exception("Can't find package name from the list: " + name_or_index)
                return
        
        self.listbox.select_clear(0, "end")
        self.listbox.select_set(index)
        self.listbox.see(index)
        
        
    
    def _perform_action(self, action):
        assert self._get_state() == "idle"
        assert self.current_package_data is not None
        data = self.current_package_data
        name = self.current_package_data["info"]["name"]
        install_args = ["install", "--no-cache-dir"] 
        if self._only_user:
            install_args.append("--user")
        
        if action == "install":
            args = install_args
            if self._get_installed_version(name) is not None:
                args.append("--upgrade")
            
            args.append(name)
        elif action == "uninstall":
            if (name in ["pip", "setuptools"]
                and not messagebox.askyesno("Really uninstall?",
                    "Package '{}' is required for installing and uninstalling other packages.\n\n".format(name)
                    + "Are you sure you want to uninstall it?")):
                return
            args = ["uninstall", "-y", name]
        elif action == "advanced":
            details = _ask_installation_details(self, data, 
                        _get_latest_stable_version(list(data["releases"].keys())))
            if details is None: # Cancel
                return
            
            version, upgrade_deps = details
            args = install_args
            if upgrade_deps:
                args.append("--upgrade")
            args.append(name + "==" + version)
        else:
            raise RuntimeError("Unknown action")
        
        proc, cmd = self._create_pip_process(args)
        title = subprocess.list2cmdline(cmd)
        
        # following call blocks
        _show_subprocess_dialog(self, proc, title)
        if action == "uninstall":
            self._show_instructions() # Make the old package go away as fast as possible
        self._start_update_list(None if action == "uninstall" else name)
        
        
    
    def _handle_install_file_click(self, event):
        if self._get_state() != "idle":
            return 
        
        filename = askopenfilename (
            filetypes = [('Package', '.whl .zip .tar.gz'), ('all files', '.*')], 
            initialdir = get_workbench().get_option("run.working_directory")
        )
        if filename: # Note that missing filename may be "" or () depending on tkinter version
            self._install_local_file(filename)
    
    def _install_local_file(self, filename):
        args = ["install"]
        if self._only_user:
            args.append("--user")
        args.append(filename)
        
        proc, cmd = self._create_pip_process(args)
        # following call blocks
        title = subprocess.list2cmdline(cmd)
        
        _, out, _ = _show_subprocess_dialog(self, proc, title)
        
        # Try to find out the name of the package we're installing
        name = None
        
        # output should include a line like this:
        # Installing collected packages: pytz, six, python-dateutil, numpy, pandas
        inst_lines = re.findall("^Installing collected packages:.*?$", out,
                                     re.MULTILINE | re.IGNORECASE)  # @UndefinedVariable
        if len(inst_lines) == 1:
            # take last element
            elements = re.split(",|:", inst_lines[0])
            name = elements[-1].strip()
        
        self._start_update_list(name)
    
    def _handle_url_click(self, event):
        url = _extract_click_text(self.info_text, event, "url")
        if url is not None:
            webbrowser.open(url)
    
    def _on_close(self, event=None):
        self.destroy()
        
    def _get_installed_version(self, name):
        for list_name in self._installed_versions:
            if self._normalize_name(name) == self._normalize_name(list_name):
                return self._installed_versions[list_name]
        
        return None

    def _prepare_env_for_pip_process(self, encoding):
        env = {}
        for name in os.environ:
            if ("python" not in name.lower() and name not in ["TK_LIBRARY", "TCL_LIBRARY"]): # skip python vars
                env[name] = os.environ[name]
                
        env["PYTHONIOENCODING"] = encoding
        env["PYTHONUNBUFFERED"] = "1"
        
        return env

    def _create_backend_process(self, args, stderr=subprocess.STDOUT):
        encoding = "UTF-8"
        
                    
        cmd = [self._get_interpreter()] + args
        
        startupinfo = None
        creationflags = 0
        if running_on_windows():
            creationflags = subprocess.CREATE_NEW_PROCESS_GROUP
            startupinfo = subprocess.STARTUPINFO()
            startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW
        
        return (subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=stderr,
                                env=self._prepare_env_for_pip_process(encoding),
                                universal_newlines=True,
                                creationflags=creationflags,
                                startupinfo=startupinfo),
                cmd)

    def _create_pip_process(self, args):
        return self._create_backend_process(["-m", "pip"] + args)
    
    def _get_interpreter(self):
        pass
    
    def _get_title(self):
        return "Manage packages for " + self._get_interpreter()
    
    def _confirm_install(self, name):
        return True
    
    def _read_only(self):
        return False

class BackendPipDialog(PipDialog):
    def _get_interpreter(self):
        return get_runner().get_interpreter_command()

    def _confirm_install(self, name):
        if name.lower().startswith("thonny"):
            return messagebox.askyesno("Confirmation", 
                                     "Looks like you are installing a Thonny-related package.\n"
                                   + "If you meant to install a Thonny plugin, then you should\n"
                                   + "close this dialog and choose 'Tools → Manage plugins...'\n"
                                   + "\n"
                                   + "Are you sure you want to install '" + name + "' here?")
        else:
            return True

    def _handle_outdated_or_missing_pip(self):
        if get_runner().using_venv():
            self._install_pip()
        else:
            self._provide_pip_install_instructions()
        
    def _read_only(self):
        return not get_runner().using_venv()

class PluginsPipDialog(PipDialog):
    def __init__(self, master):
        PipDialog.__init__(self, master, only_user=True)
    
    def _get_interpreter(self):
        return sys.executable.replace("thonny.exe", "python.exe")
    
    def _prepare_env_for_pip_process(self, encoding):
        env = PipDialog._prepare_env_for_pip_process(self, encoding)
        env["PYTHONUSERBASE"] = THONNY_USER_BASE
        return env
        
        
    def _create_widgets(self, parent):
        bg = "#ffff99"
        banner = tk.Label(parent, background=bg)
        banner.grid(row=0, column=0, sticky="nsew")
        
        banner_text = tk.Label(banner, text="NB! This dialog is for managing Thonny plug-ins and their dependencies.\n"
                                + "If you want to install packages for your own programs then close this and choose 'Tools → Manage packages...'\n"
                                + "\n"
                                + "This dialog installs packages into " + THONNY_USER_BASE + "\n"
                                + "\n"
                                + "NB! You need to restart Thonny after installing / upgrading / uninstalling a plug-in.",
                                background=bg, justify="left")
        banner_text.grid(pady=10, padx=10)
        
        PipDialog._create_widgets(self, parent)
    
    def _get_title(self):
        return "Thonny plug-ins"

    def _handle_outdated_or_missing_pip(self):
        return self._provide_pip_install_instructions()
    
    def _instructions_for_command_line_install(self):
        # System shell is not suitable without correct PYTHONUSERBASE 
        return ""
        
class DetailsDialog(tk.Toplevel):
    def __init__(self, master, package_metadata, selected_version):
        tk.Toplevel.__init__(self, master)
        self.result = None
        
        self.title("Advanced install / upgrade / downgrade")

        self.rowconfigure(0, weight=1)
        self.columnconfigure(0, weight=1)
        main_frame = ttk.Frame(self) # To get styled background
        main_frame.grid(sticky="nsew")
        main_frame.rowconfigure(0, weight=1)
        main_frame.columnconfigure(0, weight=1)
        
        version_label = ttk.Label(main_frame, text="Desired version")
        version_label.grid(row=0, column=0, columnspan=2, padx=20, pady=(15,0), sticky="w")
        
        def version_sort_key(s):
            # Trying to massage understandable versions into valid StrictVersions
            if s.replace(".", "").isnumeric(): # stable release
                s2 = s + "b999" # make it latest beta version
            elif "rc" in s:
                s2 = s.replace("rc", "b8")
            else:
                s2 = s
            try:
                return StrictVersion(s2)
            except:
                # use only numbers
                nums = re.findall(r"\d+", s)
                while len(nums) < 2:
                    nums.append("0")
                return StrictVersion(".".join(nums[:3]))
        
        version_strings = list(package_metadata["releases"].keys())
        version_strings.sort(key=version_sort_key, reverse=True)
        self.version_combo = ttk.Combobox(main_frame, values=version_strings,
                              exportselection=False)
        try:
            self.version_combo.current(version_strings.index(selected_version))
        except:
            pass
        
        self.version_combo.state(['!disabled', 'readonly'])
        self.version_combo.grid(row=1, column=0, columnspan=2, pady=(0,15),
                                padx=20, sticky="ew")
        
        
        self.update_deps_var = tk.IntVar()
        self.update_deps_var.set(0)
        self.update_deps_cb = ttk.Checkbutton(main_frame, text="Upgrade dependencies",
                                              variable=self.update_deps_var)
        self.update_deps_cb.grid(row=2, column=0, columnspan=2, padx=20, sticky="w")
        
        self.ok_button = ttk.Button(main_frame, text="Install", command=self._ok)
        self.ok_button.grid(row=3, column=0, pady=15, padx=(20, 0), sticky="se")
        self.cancel_button = ttk.Button(main_frame, text="Cancel", command=self._cancel)
        self.cancel_button.grid(row=3, column=1, pady=15, padx=(5,20), sticky="se")
        
        

        if misc_utils.running_on_mac_os():
            self.configure(background="systemSheetBackground")
        #self.resizable(height=tk.FALSE, width=tk.FALSE)
        self.transient(master)
        self.grab_set() # to make it active and modal
        self.version_combo.focus_set()
        
        
        self.bind('<Escape>', self._cancel, True)  
        self.protocol("WM_DELETE_WINDOW", self._cancel)
        
        ui_utils.center_window(self, master)
        
    
    def _ok(self, event=None):
        self.result = self.version_combo.get(), bool(self.update_deps_var.get())
        self.destroy()
    
    def _cancel(self, event=None):
        self.result = None
        self.destroy()
        
def _fetch_url_future(url, timeout=10):
    def load_url():
        with urlopen(url, timeout=timeout) as conn:
            return (conn, conn.read())
            
    executor = ThreadPoolExecutor(max_workers=1)
    return executor.submit(load_url)



def _get_latest_stable_version(version_strings):
    versions = []
    for s in version_strings:
        if s.replace(".", "").isnumeric(): # Assuming stable versions have only dots and numbers
            versions.append(LooseVersion(s)) # LooseVersion __str__ doesn't change the version string
    
    if len(versions) == 0:
        return None
        
    return str(sorted(versions)[-1])


def _show_subprocess_dialog(master, proc, title):
    dlg = SubprocessDialog(master, proc, title)
    dlg.wait_window()
    return dlg.returncode, dlg.stdout, dlg.stderr


def _ask_installation_details(master, data, selected_version):
    dlg = DetailsDialog(master, data, selected_version)
    dlg.wait_window()
    return dlg.result


def _extract_click_text(widget, event, tag):
    # http://stackoverflow.com/a/33957256/261181
    try:
        index = widget.index("@%s,%s" % (event.x, event.y))
        tag_indices = list(widget.tag_ranges(tag))
        for start, end in zip(tag_indices[0::2], tag_indices[1::2]):
            # check if the tag matches the mouse click index
            if widget.compare(start, '<=', index) and widget.compare(index, '<', end):
                return widget.get(start, end)
    except:
        logging.exception("extracting click text")
        return None


def load_plugin():
    def open_backend_pip_gui(*args):
        pg = BackendPipDialog(get_workbench())
        pg.wait_window()
    
    def open_backend_pip_gui_enabled():
        return "pip_gui" in get_runner().supported_features()

    def open_frontend_pip_gui(*args):
        pg = PluginsPipDialog(get_workbench())
        pg.wait_window()

    get_workbench().add_command("backendpipgui", "tools", "Manage packages...", open_backend_pip_gui,
                                tester=open_backend_pip_gui_enabled,
                                group=80)
    get_workbench().add_command("pluginspipgui", "tools", "Manage plug-ins...", open_frontend_pip_gui,
                                group=180)


    


# -*- coding: utf-8 -*-

import importlib
import os.path
import sys
from tkinter import ttk
import traceback

from thonny import ui_utils
from thonny.code import EditorNotebook
from thonny.common import Record, UserError
from thonny.config import try_load_configuration
from thonny.misc_utils import running_on_mac_os, running_on_linux
from thonny.ui_utils import sequence_to_accelerator, AutomaticPanedWindow, AutomaticNotebook,\
    create_tooltip, get_current_notebook_tab_widget, select_sequence
import tkinter as tk
import tkinter.font as tk_font
import tkinter.messagebox as tk_messagebox
from thonny.running import Runner
import thonny.globals
import logging
from thonny.globals import register_runner, get_runner
from thonny.config_ui import ConfigurationDialog
import pkgutil
import socket
import queue
from _thread import start_new_thread
import ast
from thonny import THONNY_USER_DIR

THONNY_PORT = 4957
SERVER_SUCCESS = "OK"
CONFIGURATION_FILE_NAME = os.path.join(THONNY_USER_DIR, "configuration.ini")
SINGLE_INSTANCE_DEFAULT = True

class Workbench(tk.Tk):
    """
    Thonny's main window and communication hub.
    
    Is responsible for:
    
        * creating the main window
        * maintaining layout (_init_containers)
        * loading plugins (_init_plugins, add_view, add_command)        
        * providing references to main components (editor_notebook and runner)
        * communication between other components (see event_generate and bind)
        * configuration services (get_option, set_option, add_defaults)
        * loading translations
        * maintaining fonts (get_font, increasing and decreasing font size)
    
    After workbench and plugins get loaded, 3 kinds of events start happening:
        
        * User events (keypresses, mouse clicks, menu selections, ...)
        * Virtual events (mostly via get_workbench().event_generate). These include:
          events reported via and dispatched by Tk event system;
          WorkbenchEvent-s, reported via and dispatched by enhanced get_workbench().event_generate.
        * Events from the background process (program output notifications, input requests,
          notifications about debugger's progress)
          
    """
    def __init__(self, server_socket=None):
        self._destroying = False
        self.initializing = True
        tk.Tk.__init__(self, className="Thonny")
        # self.tk.call("tk", "scaling", 2.0)
        tk.Tk.report_callback_exception = self._on_tk_exception
        self._event_handlers = {}
        self._images = set() # to avoid Python garbage collecting them
        self._image_mapping = {} # to allow specify different images in a theme
        self._backends = {}
        self._theme_tweaker = None
        thonny.globals.register_workbench(self)
        
        self._init_configuration()
        self._init_diagnostic_logging()
        logging.info("Loading early plugins from " + str(sys.path))
        self._load_early_plugins()
        
        self._editor_notebook = None
        self._select_theme()
        self._init_fonts()
        self._init_window()
        self._init_menu()
        
        self.title("Thonny")
        
        self._init_containers()
        
        self._init_runner()
            
        self._init_commands()
        self._load_plugins()
        
        self._update_toolbar()
        try:
            self._editor_notebook.load_startup_files()
        except:
            self.report_exception()
            
        self._editor_notebook.focus_set()
        self._try_action(self._open_views)
        
        if server_socket is not None:
            self._init_server_loop(server_socket)
        
        self.bind_class("CodeViewText", "<<CursorMove>>", self.update_title, True)
        self.bind_class("CodeViewText", "<<Modified>>", self.update_title, True)
        self.bind_class("CodeViewText", "<<TextChange>>", self.update_title, True)
        self.get_editor_notebook().bind("<<NotebookTabChanged>>", self.update_title ,True)
        
        self.initializing = False
    
    def _try_action(self, action):
        try:
            action()
        except:
            self.report_exception()
        
    def _init_configuration(self):
        self._configuration_manager = try_load_configuration(CONFIGURATION_FILE_NAME)
        self._configuration_pages = {}

        self.set_default("general.single_instance", SINGLE_INSTANCE_DEFAULT)
        self.set_default("general.expert_mode", False)
        self.set_default("debug_mode", False)

    
    def _init_diagnostic_logging(self):
        logFormatter = logging.Formatter('%(levelname)s: %(message)s')
        root_logger = logging.getLogger()
        
        log_file = os.path.join(THONNY_USER_DIR, "frontend.log")
        file_handler = logging.FileHandler(log_file, encoding="UTF-8", mode="w")
        file_handler.setFormatter(logFormatter)
        file_handler.setLevel(logging.INFO);
        root_logger.addHandler(file_handler)
        
        console_handler = logging.StreamHandler(sys.stdout)
        console_handler.setFormatter(logFormatter)
        console_handler.setLevel(logging.INFO);
        root_logger.addHandler(console_handler)
        
        root_logger.setLevel(logging.INFO)
        
        import faulthandler
        fault_out = open(os.path.join(THONNY_USER_DIR, "frontend_faults.log"), mode="w")
        faulthandler.enable(fault_out)
        
    def _init_window(self):
        
        self.set_default("layout.zoomed", False)
        self.set_default("layout.top", 15)
        self.set_default("layout.left", 150)
        self.set_default("layout.width", 700)
        self.set_default("layout.height", 650)
        self.set_default("layout.w_width", 200)
        self.set_default("layout.e_width", 200)
        self.set_default("layout.s_height", 200)
        
        # I don't actually need saved options for Full screen/maximize view,
        # but it's easier to create menu items, if I use configuration manager's variables
        self.set_default("view.full_screen", False)  
        self.set_default("view.maximize_view", False)
        
        # In order to avoid confusion set these settings to False 
        # even if they were True when Thonny was last run
        self.set_option("view.full_screen", False)
        self.set_option("view.maximize_view", False)
        
        
        self.geometry("{0}x{1}+{2}+{3}".format(self.get_option("layout.width"),
                                            self.get_option("layout.height"),
                                            self.get_option("layout.left"),
                                            self.get_option("layout.top")))
        
        if self.get_option("layout.zoomed"):
            ui_utils.set_zoomed(self, True)
        
        self.protocol("WM_DELETE_WINDOW", self._on_close)
        
        # Window icons
        window_icons = self.get_option("theme.window_icons") 
        if window_icons:
            imgs = [self.get_image(filename) for filename in window_icons]
            self.iconphoto(True, *imgs)
        elif running_on_linux() and ui_utils.get_tk_version_info() >= (8,6):
            self.iconphoto(True, self.get_image("thonny.png"))
        else:
            icon_file = os.path.join(self.get_package_dir(), "res", "thonny.ico")
            try:
                self.iconbitmap(icon_file, default=icon_file)
            except:
                try:
                    # seems to work in mac
                    self.iconbitmap(icon_file)
                except:
                    pass # TODO: try to get working in Ubuntu  
        
        self.bind("<Configure>", self._on_configure, True)
        
    def _init_menu(self):
        self.option_add('*tearOff', tk.FALSE)
        self._menubar = tk.Menu(self, **self.get_option("theme.menubar_options", {
            #"relief" : "flat",
            "activeborderwidth" : 0
        }))
        self["menu"] = self._menubar
        self._menus = {}
        self._menu_item_groups = {} # key is pair (menu_name, command_label)
        self._menu_item_testers = {} # key is pair (menu_name, command_label)
        
        # create standard menus in correct order
        self.get_menu("file", "File")
        self.get_menu("edit", "Edit")
        self.get_menu("view", "View")
        self.get_menu("run", "Run")
        self.get_menu("tools", "Tools")
        self.get_menu("help", "Help")
    
    def _load_early_plugins(self):
        """load_early_plugin can't use nor GUI neither Runner"""
        self._load_plugins("load_early_plugin")
        
    def _load_plugins(self, load_function_name="load_plugin"):
        # built-in plugins
        import thonny.plugins
        self._load_plugins_from_path(thonny.plugins.__path__, "thonny.plugins.",
                                     load_function_name=load_function_name)
        
        # 3rd party plugins from namespace package
        try:
            import thonnycontrib  # @UnresolvedImport
        except ImportError:
            # No 3rd party plugins installed
            pass
        else:
            self._load_plugins_from_path(thonnycontrib.__path__, "thonnycontrib.",
                                     load_function_name=load_function_name)
        
    def _load_plugins_from_path(self, path, prefix="", load_function_name="load_plugin"):
        for _, module_name, _ in pkgutil.iter_modules(path, prefix):
            try:
                m = importlib.import_module(module_name)
                if hasattr(m, load_function_name):
                    getattr(m, load_function_name)()
            except:
                logging.exception("Failed loading plugin '" + module_name + "'")
    
                                
    def _init_fonts(self):
        self.set_default("view.io_font_family", 
                        "Courier" if running_on_mac_os() else "Courier New")
        
        default_editor_family = "Courier New"
        families = tk_font.families()
        
        for family in ["Consolas", "Ubuntu Mono", "Menlo", "DejaVu Sans Mono"]:
            if family in families:
                default_editor_family = family
                break
        
        self.set_default("view.editor_font_family", default_editor_family)
        self.set_default("view.editor_font_size", 
                        14 if running_on_mac_os() else 11)

        default_font = tk_font.nametofont("TkDefaultFont")

        self._fonts = {
            'IOFont' : tk_font.Font(family=self.get_option("view.io_font_family")),
            'EditorFont' : tk_font.Font(family=self.get_option("view.editor_font_family")),
            'BoldEditorFont' : tk_font.Font(family=self.get_option("view.editor_font_family"),
                                            weight="bold"),
            'ItalicEditorFont' : tk_font.Font(family=self.get_option("view.editor_font_family"),
                                            slant="italic"),
            'BoldItalicEditorFont' : tk_font.Font(family=self.get_option("view.editor_font_family"),
                                            weight="bold", slant="italic"),
            'TreeviewFont' : tk_font.Font(family=default_font.cget("family"),
                                          size=default_font.cget("size"))
        }
        
        self.update_fonts()
        
    def _init_runner(self):
        try:
            runner = Runner()
            register_runner(runner)
            runner.start()
        except:
            self.report_exception("Error when initializing backend")
    
    def _init_server_loop(self, server_socket):
        """Socket will listen requests from newer Thonny instances,
        which try to delegate opening files to older instance"""
        self._requests_from_socket = queue.Queue()
        
        def server_loop():
            while True:
                logging.debug("Waiting for next client")
                (client_socket, _) = server_socket.accept()
                try:
                    self._handle_socket_request(client_socket)
                except:
                    traceback.print_exc()
        
        start_new_thread(server_loop, ())
        self._poll_socket_requests()

    def _init_commands(self):
        
        self.add_command("exit", "file", "Exit",
            self._on_close, 
            default_sequence=select_sequence("<Alt-F4>", "<Command-q>"))
        
        
        self.add_command("show_options", "tools", "Options...", self._cmd_show_options, group=180)
        self.createcommand("::tk::mac::ShowPreferences", self._cmd_show_options)
        
        self.add_command("increase_font_size", "view", "Increase font size",
            lambda: self._change_font_size(1),
            default_sequence=select_sequence("<Control-plus>", "<Command-Shift-plus>"),
            group=60)
                
        self.add_command("decrease_font_size", "view", "Decrease font size",
            lambda: self._change_font_size(-1),
            default_sequence=select_sequence("<Control-minus>", "<Command-minus>"),
            group=60)
        
        self.bind("<Control-MouseWheel>", self._cmd_zoom_with_mouse, True)
        
        self.add_command("focus_editor", "view", "Focus editor",
            self._cmd_focus_editor,
            default_sequence="<Alt-e>",
            group=70)
        
                
        self.add_command("focus_shell", "view", "Focus shell",
            self._cmd_focus_shell,
            default_sequence="<Alt-s>",
            group=70)
        
        if self.get_option("general.expert_mode"):
            
            self.add_command("toggle_maximize_view", "view", "Maximize view",
                self._cmd_toggle_maximize_view,
                flag_name="view.maximize_view",
                default_sequence=None,
                group=80)
            self.bind_class("TNotebook", "<Double-Button-1>", self._maximize_view, True)
            self.bind("<Escape>", self._unmaximize_view, True)
            
            if not running_on_mac_os():
                # TODO: approach working in Win/Linux doesn't work in mac as it should and only confuses
                self.add_command("toggle_maximize_view", "view", "Full screen",
                    self._cmd_toggle_full_screen,
                    flag_name="view.full_screen",
                    default_sequence=select_sequence("<F11>", "<Command-Shift-F>"),
                    group=80)
        
            
    def _init_containers(self):
        
        # Main frame functions as
        # - a backgroud behind padding of main_pw, without this OS X leaves white border
        # - a container to be hidden, when a view is maximized and restored when view is back home
        main_frame= ttk.Frame(self) # 
        self._main_frame = main_frame
        main_frame.grid(row=0, column=0, sticky=tk.NSEW)
        self.columnconfigure(0, weight=1)
        self.rowconfigure(0, weight=1)
        self._maximized_view = None
        
        self._toolbar = ttk.Frame(main_frame, padding=0) # TODO: height=30 ?
        self._toolbar.grid(column=0, row=0, sticky=tk.NSEW, padx=10, pady=(5,0))
        
        self.set_default("layout.main_pw_first_pane_size", 1/3)
        self.set_default("layout.main_pw_last_pane_size", 1/3)
        self._main_pw = AutomaticPanedWindow(main_frame, orient=tk.HORIZONTAL,
            first_pane_size=self.get_option("layout.main_pw_first_pane_size"),
            last_pane_size=self.get_option("layout.main_pw_last_pane_size")
        )
        
        self._main_pw.grid(column=0, row=1, sticky=tk.NSEW, padx=10, pady=10)
        main_frame.columnconfigure(0, weight=1)
        main_frame.rowconfigure(1, weight=1)
        
        self.set_default("layout.west_pw_first_pane_size", 1/3)
        self.set_default("layout.west_pw_last_pane_size", 1/3)
        self.set_default("layout.center_pw_first_pane_size", 1/3)
        self.set_default("layout.center_pw_last_pane_size", 1/3)
        self.set_default("layout.east_pw_first_pane_size", 1/3)
        self.set_default("layout.east_pw_last_pane_size", 1/3)
        
        self._west_pw = AutomaticPanedWindow(self._main_pw, 1, orient=tk.VERTICAL,
            first_pane_size=self.get_option("layout.west_pw_first_pane_size"),
            last_pane_size=self.get_option("layout.west_pw_last_pane_size")
        )
        self._center_pw = AutomaticPanedWindow(self._main_pw, 2, orient=tk.VERTICAL,
            first_pane_size=self.get_option("layout.center_pw_first_pane_size"),
            last_pane_size=self.get_option("layout.center_pw_last_pane_size")
        )
        self._east_pw = AutomaticPanedWindow(self._main_pw, 3, orient=tk.VERTICAL,
            first_pane_size=self.get_option("layout.east_pw_first_pane_size"),
            last_pane_size=self.get_option("layout.east_pw_last_pane_size")
        )
        
        self._view_records = {}
        self._view_notebooks = {
            'nw' : AutomaticNotebook(self._west_pw, 1),
            'w'  : AutomaticNotebook(self._west_pw, 2),
            'sw' : AutomaticNotebook(self._west_pw, 3),
            
            's'  : AutomaticNotebook(self._center_pw, 3),
            
            'ne' : AutomaticNotebook(self._east_pw, 1),
            'e'  : AutomaticNotebook(self._east_pw, 2),
            'se' : AutomaticNotebook(self._east_pw, 3),
        }
        
        for nb_name in self._view_notebooks:
            self.set_default("layout.notebook_" + nb_name + "_visible_view", None)

        self._editor_notebook = EditorNotebook(self._center_pw)
        self._editor_notebook.position_key = 1
        self._center_pw.insert("auto", self._editor_notebook)

    def _select_theme(self):
        style = ttk.Style()
        
        preferred_theme = self.get_option("theme.preferred_theme")
        
        if preferred_theme in style.theme_names():
            style.theme_use(preferred_theme)
        elif 'xpnative' in style.theme_names():
            # in Win7 'xpnative' gives better scrollbars than 'vista'
            style.theme_use('xpnative') 
        elif 'vista' in style.theme_names():
            style.theme_use('vista')
        elif 'clam' in style.theme_names():
            style.theme_use('clam')
        
        if self._theme_tweaker is not None:
            self._theme_tweaker()

        
    def add_command(self, command_id, menu_name, command_label, handler,
                    tester=None,
                    default_sequence=None,
                    flag_name=None,
                    skip_sequence_binding=False,
                    accelerator=None,
                    group=99,
                    position_in_group="end",
                    image_filename=None,
                    include_in_toolbar=False,
                    bell_when_denied=True):
        """Adds an item to specified menu.
        
        Args:
            menu_name: Name of the menu the command should appear in.
                Standard menu names are "file", "edit", "run", "view", "help".
                If a menu with given name doesn't exist, then new menu is created
                (with label=name).
            command_label: Label for this command
            handler: Function to be called when the command is invoked. 
                Should be callable with one argument (the event or None).
            tester: Function to be called for determining if command is available or not.
                Should be callable with one argument (the event or None).
                Should return True or False.
                If None then command is assumed to be always available.
            default_sequence: Default shortcut (Tk style)
            flag_name: Used for toggle commands. Indicates the name of the boolean option.
            group: Used for grouping related commands together. Value should be int. 
                Groups with smaller numbers appear before.
        
        Returns:
            None
        """     
        
        def dispatch(event=None):
            if not tester or tester():
                denied = False
                handler()
            else:
                denied = True
                logging.debug("Command '" + command_id + "' execution denied")
                if bell_when_denied:
                    self.bell()
                
            self.event_generate("Command", command_id=command_id, denied=denied)
        
        sequence_option_name = "shortcuts." + command_id
        self.set_default(sequence_option_name, default_sequence)
        sequence = self.get_option(sequence_option_name) 
        
        if sequence and not skip_sequence_binding:
            self.bind_all(sequence, dispatch, True)
        
        
        def dispatch_from_menu():
            # I don't like that Tk menu toggles checbutton variable
            # automatically before calling the handler.
            # So I revert the toggle before calling the actual handler.
            # This way the handler doesn't have to worry whether it
            # needs to toggle the variable or not, and it can choose to 
            # decline the toggle.
            if flag_name is not None:
                var = self.get_variable(flag_name)
                var.set(not var.get())
                
            dispatch(None)
        
        if image_filename:
            image = self.get_image(image_filename)
        else:
            image = None
        
        if image and self.get_option("theme.icons_in_menus", True):
            menu_image = image
        elif flag_name: 
            # no image or black next to a checkbox
            menu_image = None
        else:
            menu_image = self.get_image ("16x16_blank.gif")
        
        if not accelerator and sequence:
            accelerator = sequence_to_accelerator(sequence)
        
        menu = self.get_menu(menu_name)
        menu.insert(
            self._find_location_for_menu_item(menu_name, command_label, group, position_in_group),
            "checkbutton" if flag_name else "command",
            label=command_label,
            accelerator=accelerator,
            image=menu_image, 
            compound=tk.LEFT,
            variable=self.get_variable(flag_name) if flag_name else None,
            command=dispatch_from_menu)
        
        # remember the details that can't be stored in Tkinter objects
        self._menu_item_groups[(menu_name, command_label)] = group
        self._menu_item_testers[(menu_name, command_label)] = tester
        
        if include_in_toolbar:
            toolbar_group = self._get_menu_index(menu) * 100 + group
            self._add_toolbar_button(image, command_label, accelerator, handler, tester,
                toolbar_group)
        
    
    def add_view(self, class_, label, default_location,
                visible_by_default=False,
                default_position_key=None):
        """Adds item to "View" menu for showing/hiding given view. 
        
        Args:
            view_class: Class or constructor for view. Should be callable with single
                argument (the master of the view)
            label: Label of the view tab
            location: Location descriptor. Can be "nw", "sw", "s", "se", "ne"
        
        Returns: None        
        """
        view_id = class_.__name__
        if default_position_key == None:
            default_position_key = label
        
        self.set_default("view." + view_id + ".visible" , visible_by_default)
        self.set_default("view." + view_id + ".location", default_location)
        self.set_default("view." + view_id + ".position_key", default_position_key)
        
        self._view_records[view_id] = {
            "class" : class_,
            "label" : label,
            "location" : self.get_option("view." + view_id + ".location"),
            "position_key" : self.get_option("view." + view_id + ".position_key")
        }
        
        visibility_flag = self.get_variable("view." + view_id + ".visible")
        
        # handler
        def toggle_view_visibility():
            if visibility_flag.get():
                self.hide_view(view_id)
            else:
                self.show_view(view_id, True)
        
        self.add_command("toggle_" + view_id,
            menu_name="view",
            command_label=label,
            handler=toggle_view_visibility,
            flag_name="view." + view_id + ".visible",
            group=10,
            position_in_group="alphabetic")
        
        if visibility_flag.get():
            self.show_view(view_id, False)
    
    def add_configuration_page(self, title, page_class):
        self._configuration_pages[title] = page_class
    
    def map_image(self, original_image, new_image):
        self._image_mapping[original_image] = new_image
    
    def add_backend(self, descriptor, proxy_class):
        self._backends[descriptor] = proxy_class
    
    def get_backends(self):
        return self._backends
    
    def get_option(self, name, default=None):
        return self._configuration_manager.get_option(name, default)
    
    def set_option(self, name, value):
        self._configuration_manager.set_option(name, value)
    
    def set_theme_tweaker(self, fun):
        self._theme_tweaker = fun
    
    def set_default(self, name, default_value):
        """Registers a new option.
        
        If the name contains a period, then the part left to the (first) period
        will become the section of the option and rest will become name under that 
        section.
        
        If the name doesn't contain a period, then it will be added under section 
        "general".
        """
        self._configuration_manager.set_default(name, default_value)
    
    def get_variable(self, name):
        return self._configuration_manager.get_variable(name)
    
    def get_font(self, name):
        """
        Supported names are EditorFont and BoldEditorFont
        """
        return self._fonts[name]
    
    
    def get_menu(self, name, label=None):
        """Gives the menu with given name. Creates if not created yet.
        
        Args:
            name: meant to be used as not translatable menu name
            label: translated label, used only when menu with given name doesn't exist yet
        """
        if name not in self._menus:
            menu = tk.Menu(self._menubar, self.get_option("theme.menu_options", {
                #"relief" : "flat",
                #"activeborderwidth" : 0
            }))
            menu["postcommand"] = lambda: self._update_menu(menu, name)
            self._menubar.add_cascade(label=label if label else name, menu=menu)
            
            self._menus[name] = menu
            if label:
                self._menus[label] = menu
                
        return self._menus[name]
    
    def get_view(self, view_id, create=True):
        if "instance" not in self._view_records[view_id]:
            if not create:
                return None
            class_ = self._view_records[view_id]["class"]
            location = self._view_records[view_id]["location"]
            master = self._view_notebooks[location]
            
            # create the view
            view = class_(self) # View's master is workbench to allow making it maximized
            view.position_key = self._view_records[view_id]["position_key"]
            self._view_records[view_id]["instance"] = view

            # create the view home_widget to be added into notebook
            view.home_widget = ttk.Frame(master) 
            view.home_widget.columnconfigure(0, weight=1)
            view.home_widget.rowconfigure(0, weight=1)
            view.home_widget.maximizable_widget = view
            if hasattr(view, "position_key"):
                view.home_widget.position_key = view.position_key
            
            # initially the view will be in it's home_widget
            view.grid(row=0, column=0, sticky=tk.NSEW, in_=view.home_widget)
            view.hidden = True
            
        return self._view_records[view_id]["instance"]
    
    def get_current_editor(self):
        return self._editor_notebook.get_current_editor()
    
    def get_editor_notebook(self):
        return self._editor_notebook
    
    def get_package_dir(self):
        """Returns thonny package directory"""
        return os.path.dirname(sys.modules["thonny"].__file__)
    
    def get_image(self, filename, tk_name=None):
        
        if filename in self._image_mapping:
            filename = self._image_mapping[filename]
        
        # if path is relative then interpret it as living in res folder
        if not os.path.isabs(filename):
            filename = os.path.join(self.get_package_dir(), "res", filename)
            
        img = tk.PhotoImage(tk_name, file=filename)
        self._images.add(img)
        return img
                      
    def show_view(self, view_id, set_focus=True):
        """View must be already registered.
        
        Args:
            view_id: View class name 
            without package name (eg. 'ShellView') """

        # NB! Don't forget that view.home_widget is added to notebook, not view directly
        # get or create
        view = self.get_view(view_id)
        notebook = view.home_widget.master
        
        if hasattr(view, "before_show") and view.before_show() == False:
            return False
            
        if view.hidden:
            notebook.insert("auto", view.home_widget, text=self._view_records[view_id]["label"])
            view.hidden = False
        
        # switch to the tab
        notebook.select(view.home_widget)
        
        # add focus
        if set_focus:
            view.focus_set()
        
        self.set_option("view." + view_id + ".visible", True)
        self.event_generate("ShowView", view=view, view_id=view_id)
        return view
    
    def hide_view(self, view_id):
        # NB! Don't forget that view.home_widget is added to notebook, not view directly
        
        if "instance" in self._view_records[view_id]:
            # TODO: handle the case, when view is maximized
            view = self._view_records[view_id]["instance"]
            
            if hasattr(view, "before_hide") and view.before_hide() == False:
                return False
            
            view.home_widget.master.forget(view.home_widget)
            
            self.set_option("view." + view_id + ".visible", False)
            
            self.event_generate("HideView", view=view, view_id=view_id)
            view.hidden = True

        

    def event_generate(self, sequence, **kwargs):
        """Uses custom event handling when sequence doesn't start with <.
        In this case arbitrary attributes can be added to the event.
        Otherwise forwards the call to Tk's event_generate"""
        if sequence.startswith("<"):
            tk.Tk.event_generate(self, sequence, **kwargs)
        else:
            if sequence in self._event_handlers:
                for handler in self._event_handlers[sequence]:
                    try:
                        # Yes, I'm creating separate event object for each handler
                        # so that they can't misuse the mutabilty
                        event = WorkbenchEvent(sequence, **kwargs)
                        handler(event)
                    except:
                        self.report_exception("Problem when handling '" + sequence + "'")
                
    def bind(self, sequence, func, add=None):
        """Uses custom event handling when sequence doesn't start with <.
        Otherwise forwards the call to Tk's bind"""
        
        if not add:
            logging.warning("Workbench.bind({}, ..., add={}) -- did you really want to replace existing bindings?".format(sequence, add))
        
        if sequence.startswith("<"):
            tk.Tk.bind(self, sequence, func, add)
        else:
            if sequence not in self._event_handlers or not add:
                self._event_handlers[sequence] = set()
                
            self._event_handlers[sequence].add(func)

    def unbind(self, sequence, funcid=None):
        if sequence.startswith("<"):
            tk.Tk.unbind(self, sequence, funcid=funcid)
        else:
            if (sequence in self._event_handlers 
                and funcid in self._event_handlers[sequence]):
                self._event_handlers[sequence].remove(funcid)
                

    def in_heap_mode(self):
        # TODO: add a separate command for enabling the heap mode 
        # untie the mode from HeapView
        
        return (self._configuration_manager.has_option("view.HeapView.visible")
            and self.get_option("view.HeapView.visible"))
    
    def update_fonts(self):
        editor_font_size = self._guard_font_size(self.get_option("view.editor_font_size"))
        editor_font_family = self.get_option("view.editor_font_family")
        io_font_family = self.get_option("view.io_font_family")
        
        self.get_font("IOFont").configure(family=io_font_family,
                                          size=min(editor_font_size - 2,
                                                   int(editor_font_size * 0.8 + 3)))
        self.get_font("EditorFont").configure(family=editor_font_family,
                                              size=editor_font_size)
        self.get_font("BoldEditorFont").configure(family=editor_font_family,
                                                  size=editor_font_size)
        self.get_font("ItalicEditorFont").configure(family=editor_font_family,
                                                  size=editor_font_size)
        self.get_font("BoldItalicEditorFont").configure(family=editor_font_family,
                                                  size=editor_font_size)
        
        
        style = ttk.Style()
        if running_on_mac_os():
            treeview_font_size = int(editor_font_size * 0.7 + 4)
            rowheight = int(treeview_font_size*1.2 + 4 )
        else:
            treeview_font_size = int(editor_font_size * 0.7 + 2)
            rowheight = int(treeview_font_size * 2.0 + 6)
            
        self.get_font("TreeviewFont").configure(size=treeview_font_size)
        style.configure("Treeview", rowheight=rowheight)
        
        if self._editor_notebook is not None:
            self._editor_notebook.update_appearance()
        
    
    def _get_menu_index(self, menu):
        for i in range(len(self._menubar.winfo_children())):
            if menu == self._menubar.winfo_children()[i]:
                return i
        else:
            return None
    
    def _add_toolbar_button(self, image, command_label, accelerator, handler, 
                            tester, toolbar_group):
        
        slaves = self._toolbar.grid_slaves(0, toolbar_group)
        if len(slaves) == 0:
            group_frame = ttk.Frame(self._toolbar)
            group_frame.grid(row=0, column=toolbar_group, padx=(0, 10))
        else:
            group_frame = slaves[0]
        
        button = ttk.Button(group_frame, 
                         command=handler, 
                         image=image, 
                         style="Toolbutton", # TODO: does this cause problems in some Macs?
                         state=tk.NORMAL
                         )
        button.pack(side=tk.LEFT)
        button.tester = tester 
        tooltip_text = command_label
        if accelerator and self.get_option("theme.shortcuts_in_tooltips", True):
            tooltip_text += " (" + accelerator + ")"
        create_tooltip(button, tooltip_text,
                       **self.get_option("theme.tooltip_options", {'padx':3, 'pady':1})
                       )
        
    def _update_toolbar(self):
        for group_frame in self._toolbar.grid_slaves(0):
            for button in group_frame.pack_slaves():
                if button.tester and not button.tester():
                    button["state"] = tk.DISABLED
                else:
                    button["state"] = tk.NORMAL
        
        self.after(300, self._update_toolbar)
            
    
    def _cmd_zoom_with_mouse(self, event):
        if event.delta > 0:
            self._change_font_size(1)
        else:
            self._change_font_size(-1)
    
    def _change_font_size(self, delta):
        
        if delta != 0:
            editor_font_size = self.get_option("view.editor_font_size")
            editor_font_size += delta
            self.set_option("view.editor_font_size", self._guard_font_size(editor_font_size))
            self.update_fonts()
    
    def _guard_font_size(self, size):
        # https://bitbucket.org/plas/thonny/issues/164/negative-font-size-crashes-thonny
        MIN_SIZE = 4
        MAX_SIZE = 200
        if size < MIN_SIZE:
            return MIN_SIZE
        elif size > MAX_SIZE:
            return MAX_SIZE
        else:
            return size
        
        
    
    def _check_update_window_width(self, delta):
        if not ui_utils.get_zoomed(self):
            self.update_idletasks()
            # TODO: shift to left if right edge goes away from screen
            # TODO: check with screen width
            new_geometry = "{0}x{1}+{2}+{3}".format(self.winfo_width() + delta,
                                                   self.winfo_height(),
                                                   self.winfo_x(), self.winfo_y())
            
            self.geometry(new_geometry)
            
    
    def _maximize_view(self, event=None):
        if self._maximized_view is not None:
            return
        
        # find the widget that can be relocated
        widget = self.focus_get()
        if isinstance(widget, EditorNotebook) or isinstance(widget, AutomaticNotebook):
            current_tab = get_current_notebook_tab_widget(widget)
            if current_tab is None:
                return
            
            if not hasattr(current_tab, "maximizable_widget"):
                return
            
            widget = current_tab.maximizable_widget
        
        while widget is not None:
            if hasattr(widget, "home_widget"):
                # if widget is view, then widget.master is workbench
                widget.grid(row=0, column=0, sticky=tk.NSEW, in_=widget.master)
                # hide main_frame
                self._main_frame.grid_forget()
                self._maximized_view = widget
                self.get_variable("view.maximize_view").set(True)
                break
            else:
                widget = widget.master
    
    def _unmaximize_view(self, event=None):
        if self._maximized_view is None:
            return
        
        # restore main_frame
        self._main_frame.grid(row=0, column=0, sticky=tk.NSEW, in_=self)
        # put the maximized view back to its home_widget
        self._maximized_view.grid(row=0, column=0, sticky=tk.NSEW, in_=self._maximized_view.home_widget)
        self._maximized_view = None
        self.get_variable("view.maximize_view").set(False)
    
    def _cmd_show_options(self):
        dlg = ConfigurationDialog(self, self._configuration_pages)
        dlg.focus_set()
        dlg.transient(self)
        dlg.grab_set()
        self.wait_window(dlg)
    
    def _cmd_focus_editor(self):
        self._editor_notebook.focus_set()
    
    def _cmd_focus_shell(self):
        self.show_view("ShellView", True)
    
    def _cmd_toggle_full_screen(self):
        var = self.get_variable("view.full_screen")
        var.set(not var.get())
        self.attributes("-fullscreen", var.get())
    
    def _cmd_toggle_maximize_view(self):
        if self._maximized_view is not None:
            self._unmaximize_view()
        else:
            self._maximize_view()
            
            
    
    def _update_menu(self, menu, menu_name):
        if menu.index("end") == None:
            return
        
        for i in range(menu.index("end")+1):
            item_data = menu.entryconfigure(i)
            if "label" in item_data:
                command_label = menu.entrycget(i, "label")
                tester = self._menu_item_testers[(menu_name, command_label)]

                if tester and not tester():
                    menu.entryconfigure(i, state=tk.DISABLED)
                else:
                    menu.entryconfigure(i, state=tk.ACTIVE)   
                    
                    
        
    
    def _find_location_for_menu_item(self, menu_name, command_label, group,
            position_in_group="end"):        
        
        menu = self.get_menu(menu_name)
        
        if menu.index("end") == None: # menu is empty
            return "end"
        
        this_group_exists = False
        for i in range(0, menu.index("end")+1):
            data = menu.entryconfigure(i)
            if "label" in data:
                # it's a command, not separator
                sibling_label = menu.entrycget(i, "label")
                sibling_group = self._menu_item_groups[(menu_name, sibling_label)]

                if sibling_group == group:
                    this_group_exists = True
                    if position_in_group == "alphabetic" and sibling_label > command_label:
                        return i
                    
                if sibling_group > group:
                    assert not this_group_exists # otherwise we would have found the ending separator
                    menu.insert_separator(i)
                    return i
            else:
                # We found a separator
                if this_group_exists: 
                    # it must be the ending separator for this group
                    return i
                
        else:
            # no group was bigger, ie. this should go to the end
            if not this_group_exists:
                menu.add_separator()
                
            return "end"

    def _handle_socket_request(self, client_socket):
        """runs in separate thread"""
        # read the request
        data = bytes()
        while True:
            new_data = client_socket.recv(1024)
            if len(new_data) > 0:
                data += new_data
            else:
                break
        
        self._requests_from_socket.put(data)
        
        # respond OK
        client_socket.sendall(SERVER_SUCCESS.encode(encoding='utf-8'))
        client_socket.shutdown(socket.SHUT_WR)
        print("AFTER NEW REQUEST", client_socket)
    
    def _poll_socket_requests(self):
        """runs in gui thread"""
        try:
            while not self._requests_from_socket.empty():
                data = self._requests_from_socket.get()
                args = ast.literal_eval(data.decode("UTF-8"))
                assert isinstance(args, list)
                for filename in args:
                    if os.path.exists(filename):
                        self._editor_notebook.show_file(filename)
                        
                self.become_topmost_window()
        finally:
            self.after(50, self._poll_socket_requests)

    def _on_close(self):
        if not self._editor_notebook.check_allow_closing():
            return
        
        try:
            self._save_layout()
            #ui_utils.delete_images()
            self.event_generate("WorkbenchClose")
        except:
            self.report_exception()

        self.destroy()
    
    def focus_get(self):
        try:
            return tk.Tk.focus_get(self)
        except:
            # This may give error in Ubuntu
            return None
    
    def destroy(self):
        try:
            self._destroying = True
            tk.Tk.destroy(self)
        except tk.TclError:
            logging.exception("Error while destroying workbench")
        finally:
            runner = get_runner()
            if runner != None:
                runner.kill_backend()
    
    def _on_configure(self, event):
        # called when window is moved or resized
        if (hasattr(self, "_maximized_view") # configure may happen before the attribute is defined 
            and self._maximized_view):
            # grid again, otherwise it acts weird
            self._maximized_view.grid(row=0, column=0, sticky=tk.NSEW, in_=self._maximized_view.master)
    
    def _on_tk_exception(self, exc, val, tb):
        # copied from tkinter.Tk.report_callback_exception with modifications
        # see http://bugs.python.org/issue22384
        sys.last_type = exc
        sys.last_value = val
        sys.last_traceback = tb
        self.report_exception()
    
    def report_exception(self, title="Internal error"):
        logging.exception(title)
        if tk._default_root and not self._destroying:
            (typ, value, _) = sys.exc_info()
            if issubclass(typ, UserError):
                msg = str(value)
            else:
                msg = traceback.format_exc()
            tk_messagebox.showerror(title, msg)
    
    def _open_views(self):
        for nb_name in self._view_notebooks:
            view_name = self.get_option("layout.notebook_" + nb_name + "_visible_view")
            if view_name != None:
                self.show_view(view_name)
                
        
        
    def _save_layout(self):
        self.update_idletasks()
        
        self.set_option("layout.zoomed", ui_utils.get_zoomed(self))
        
        # each AutomaticPanedWindow remember it's splits for both 2 and 3 panes
        self.set_option("layout.main_pw_first_pane_size", self._main_pw.first_pane_size)
        self.set_option("layout.main_pw_last_pane_size", self._main_pw.last_pane_size)
        self.set_option("layout.east_pw_first_pane_size", self._east_pw.first_pane_size)
        self.set_option("layout.east_pw_last_pane_size", self._east_pw.last_pane_size)
        self.set_option("layout.center_pw_last_pane_size", self._center_pw.last_pane_size)
        self.set_option("layout.west_pw_first_pane_size", self._west_pw.first_pane_size)
        self.set_option("layout.west_pw_last_pane_size", self._west_pw.last_pane_size)
        
        for nb_name in self._view_notebooks:
            widget = self._view_notebooks[nb_name].get_visible_child()
            if hasattr(widget, "maximizable_widget"):
                view = widget.maximizable_widget
                view_name = type(view).__name__
                self.set_option("layout.notebook_" + nb_name + "_visible_view", view_name)
            else:
                self.set_option("layout.notebook_" + nb_name + "_visible_view", None)
        
        if not ui_utils.get_zoomed(self):
            self.set_option("layout.top", self.winfo_y())
            self.set_option("layout.left", self.winfo_x())
            self.set_option("layout.width", self.winfo_width())
            self.set_option("layout.height", self.winfo_height())
        
        self._configuration_manager.save()
    
    #def focus_set(self):
    #    tk.Tk.focus_set(self)
    #    self._editor_notebook.focus_set()
    
    def update_title(self, event=None):
        editor = self.get_editor_notebook().get_current_editor()
        title_text = "Thonny"
        if editor != None:
            title_text += "  -  " + editor.get_long_description()
            
        self.title(title_text)
    
    def become_topmost_window(self):
        # Looks like at least on Windows all following is required for the window to get focus
        # (deiconify, ..., iconify, deiconify)
        self.deiconify()
        self.attributes('-topmost', True)
        self.after_idle(self.attributes, '-topmost', False)
        self.lift()
        
        if not running_on_linux():
            # http://stackoverflow.com/a/13867710/261181
            self.iconify()
            self.deiconify()
        
        editor = self.get_current_editor()
        if editor is not None:
            # This method is meant to be called when new file is opened, so it's safe to 
            # send the focus to the editor
            editor.focus_set()
        else:
            self.focus_set()
        

class WorkbenchEvent(Record):
    def __init__(self, sequence, **kwargs):
        Record.__init__(self, **kwargs)
        self.sequence = sequence




import tkinter as tk
import os.path
from tkinter import filedialog
from tkinter import ttk

from thonny.config_ui import ConfigurationPage
from thonny.globals import get_workbench, get_runner
from thonny.ui_utils import create_string_var
from thonny.misc_utils import running_on_windows
from thonny.running import parse_configuration


class InterpreterConfigurationPage(ConfigurationPage):
    
    def __init__(self, master):
        ConfigurationPage.__init__(self, master)
        
        self._configuration_variable = create_string_var(
            get_workbench().get_option("run.backend_configuration"))
        
        entry_label = ttk.Label(self, text="Which interpreter to use for running programs?")
        entry_label.grid(row=0, column=0, columnspan=2, sticky=tk.W)
        
        self._entry = ttk.Combobox(self,
                              exportselection=False,
                              textvariable=self._configuration_variable,
                              values=self._get_configurations())
        
        self._entry.grid(row=1, column=0, columnspan=2, sticky=tk.NSEW)
        self._entry.state(['!disabled', 'readonly'])
        
        another_label = ttk.Label(self, text="Your interpreter isn't in the list?")
        another_label.grid(row=2, column=0, columnspan=2, sticky=tk.W, pady=(10,0))
        self._select_button = ttk.Button(self,
                                         text="Locate another executable "
                                         + ("(python.exe) ..." if running_on_windows() else "(python3) ...")
                                         + "\nNB! Thonny only supports Python 3.4 and later",
                                         command=self._select_executable)
        
        self._select_button.grid(row=3, column=0, columnspan=2, sticky=tk.NSEW)
        
        self.columnconfigure(0, weight=1)
        self.columnconfigure(1, weight=1)
    
    def _get_configurations(self):
        result = []
        
        backends = get_workbench().get_backends()
        for backend_name in sorted(backends.keys()):
            backend_class = backends[backend_name]
            for configuration_option in backend_class.get_configuration_options():
                if configuration_option is None or configuration_option == "":
                    result.append(backend_name)
                else:
                    result.append("%s (%s)" % (backend_name, configuration_option))
        
        return result
    
    def _select_executable(self):
        if running_on_windows():
            options = {"filetypes" : [('Exe-files', '.exe'), ('all files', '.*')]}
        else:
            options = {}

        # TODO: get dir of current interpreter
            
        filename = filedialog.askopenfilename(**options)
        
        if filename:
            self._configuration_variable.set("Python (%s)" % os.path.realpath(filename))
    
    
    def apply(self):
        if not self._configuration_variable.modified:
            return
        
        configuration = self._configuration_variable.get()
        get_workbench().set_option("run.backend_configuration", configuration)
        
        backend, configuration_option = parse_configuration(configuration)
        if backend == "Python":
            # Store used CPython interpreters
            interpreter = configuration_option
            used_interpreters = get_workbench().get_option("run.used_interpreters")
            if interpreter not in used_interpreters:
                used_interpreters.append(configuration)
            get_workbench().set_option("run.used_interpreters", used_interpreters)
        
        get_runner().reset_backend()
        
    

def load_plugin():
    get_workbench().add_configuration_page("Interpreter", InterpreterConfigurationPage)
